import pytest

import cellrank as cr
from anndata import AnnData
from cellrank.tl.kernels import VelocityKernel, ConnectivityKernel
from cellrank.tl._constants import (
    Direction,
    DirPrefix,
    AbsProbKey,
    TermStatesKey,
    _probs,
    _colors,
    _lin_names,
    _transition,
)
from cellrank.tl.estimators._constants import P

import numpy as np
from pandas.api.types import is_categorical_dtype


def _assert_has_all_keys(adata: AnnData, direction: Direction):
    assert _transition(direction) in adata.obsp.keys()
    # check if it's not a dummy transition matrix
    assert not np.all(np.isclose(np.diag(adata.obsp[_transition(direction)].A), 1.0))
    assert f"{_transition(direction)}_params" in adata.uns.keys()

    if direction == Direction.FORWARD:
        assert str(AbsProbKey.FORWARD) in adata.obsm
        assert isinstance(adata.obsm[str(AbsProbKey.FORWARD)], cr.tl.Lineage)

        assert _colors(AbsProbKey.FORWARD) in adata.uns.keys()
        assert _lin_names(AbsProbKey.FORWARD) in adata.uns.keys()

        assert str(TermStatesKey.FORWARD) in adata.obs
        assert is_categorical_dtype(adata.obs[str(TermStatesKey.FORWARD)])

        assert _probs(TermStatesKey.FORWARD) in adata.obs

        # check the correlations with all lineages have been computed
        lin_probs = adata.obsm[str(AbsProbKey.FORWARD)]
        np.in1d(
            [f"{str(DirPrefix.FORWARD)} {key}" for key in lin_probs.names],
            adata.var.keys(),
        ).all()

    else:
        assert str(AbsProbKey.BACKWARD) in adata.obsm
        assert isinstance(adata.obsm[str(AbsProbKey.BACKWARD)], cr.tl.Lineage)

        assert _colors(AbsProbKey.BACKWARD) in adata.uns.keys()
        assert _lin_names(AbsProbKey.BACKWARD) in adata.uns.keys()

        assert str(TermStatesKey.BACKWARD) in adata.obs
        assert is_categorical_dtype(adata.obs[str(TermStatesKey.BACKWARD)])

        assert _probs(TermStatesKey.BACKWARD) in adata.obs

        # check the correlations with all lineages have been computed
        lin_probs = adata.obsm[str(AbsProbKey.BACKWARD)]
        np.in1d(
            [f"{str(DirPrefix.BACKWARD)} {key}" for key in lin_probs.names],
            adata.var.keys(),
        ).all()


class TestHighLevelPipeline:
    def test_plot_states_not_computed(self, adata: AnnData):
        with pytest.raises(RuntimeError):
            cr.pl.initial_states(adata)
        with pytest.raises(RuntimeError):
            cr.pl.terminal_states(adata)

    def test_write_transition_matrix(self, adata: AnnData):
        cr.tl.transition_matrix(adata, key="foo")

        assert "foo" in adata.obsp
        np.testing.assert_allclose(adata.obsp["foo"].A.sum(1), 1.0)
        assert "foo_params" in adata.uns

    def test_states_no_precomputed_transition_matrix(self, adata: AnnData):
        cr.tl.terminal_states(adata, key="foo")

        assert str(_transition(Direction.FORWARD)) in adata.obsp

    def test_states_use_precomputed_transition_matrix(self, adata: AnnData):
        cr.tl.transition_matrix(adata, key="foo")
        obsp_keys = set(adata.obsp.keys())

        cr.tl.terminal_states(adata, key="foo")

        assert obsp_keys == set(adata.obsp.keys())

    def test_fwd_pipeline_cflare(self, adata: AnnData):
        cr.tl.terminal_states(
            adata,
            estimator=cr.tl.estimators.CFLARE,
            cluster_key="clusters",
            method="kmeans",
            show_plots=True,
        )
        cr.pl.terminal_states(adata)
        cr.tl.lineages(adata)
        cr.pl.lineages(adata)
        cr.tl.lineage_drivers(adata, use_raw=False)

        ln = adata.obsm[str(AbsProbKey.FORWARD)].names[0]
        cr.pl.lineage_drivers(adata, ln, use_raw=False, backward=False)

        _assert_has_all_keys(adata, Direction.FORWARD)

    def test_fwd_pipeline_invalid_raw_requested(self, adata: AnnData):
        cr.tl.terminal_states(
            adata,
            estimator=cr.tl.estimators.CFLARE,
            cluster_key="clusters",
            method="kmeans",
            show_plots=True,
        )
        cr.pl.terminal_states(adata)
        cr.tl.lineages(adata)
        cr.pl.lineages(adata)
        cr.tl.lineage_drivers(adata, use_raw=False)

        ln = adata.obsm[str(AbsProbKey.FORWARD)].names[0]
        with pytest.raises(RuntimeError):
            cr.pl.lineage_drivers(adata, ln, use_raw=True, backward=False)

    def test_bwd_pipeline_cflare(self, adata: AnnData):
        cr.tl.initial_states(
            adata,
            estimator=cr.tl.estimators.CFLARE,
            cluster_key="clusters",
            method="louvain",
            show_plots=True,
        )
        cr.pl.initial_states(adata)
        cr.tl.lineages(adata, backward=True)
        cr.pl.lineages(adata, backward=True)
        cr.tl.lineage_drivers(adata, use_raw=False, backward=True)

        ln = adata.obsm[str(AbsProbKey.BACKWARD)].names[0]
        cr.pl.lineage_drivers(adata, ln, use_raw=False, backward=True)

        _assert_has_all_keys(adata, Direction.BACKWARD)

    def test_fwd_pipeline_gpcca(self, adata: AnnData):
        cr.tl.terminal_states(
            adata,
            estimator=cr.tl.estimators.GPCCA,
            cluster_key="clusters",
            method="brandts",
            show_plots=True,
        )
        cr.pl.terminal_states(adata)
        cr.tl.lineages(adata)
        cr.pl.lineages(adata)
        cr.tl.lineage_drivers(adata, use_raw=False)
        ln = adata.obsm[str(AbsProbKey.FORWARD)].names[0]
        cr.pl.lineage_drivers(adata, ln, use_raw=False, backward=False)

        _assert_has_all_keys(adata, Direction.FORWARD)

    def test_fwd_pipeline_gpcca_invalid_raw_requested(self, adata: AnnData):
        cr.tl.terminal_states(
            adata,
            estimator=cr.tl.estimators.GPCCA,
            cluster_key="clusters",
            method="brandts",
            show_plots=True,
        )
        cr.pl.terminal_states(adata)
        cr.tl.lineages(adata)
        cr.pl.lineages(adata)
        cr.tl.lineage_drivers(adata, use_raw=False)
        ln = adata.obsm[str(AbsProbKey.FORWARD)].names[0]
        with pytest.raises(RuntimeError):
            cr.pl.lineage_drivers(adata, ln, use_raw=True, backward=False)

    def test_bwd_pipeline_gpcca(self, adata: AnnData):
        cr.tl.initial_states(
            adata,
            estimator=cr.tl.estimators.GPCCA,
            cluster_key="clusters",
            method="brandts",
            show_plots=True,
        )
        cr.pl.initial_states(adata)
        cr.tl.lineages(adata, backward=True)
        cr.pl.lineages(adata, backward=True)
        cr.tl.lineage_drivers(adata, use_raw=False, backward=True)
        ln = adata.obsm[str(AbsProbKey.BACKWARD)].names[0]
        cr.pl.lineage_drivers(adata, ln, use_raw=False, backward=True)

        _assert_has_all_keys(adata, Direction.BACKWARD)

    def test_multiple_read_write_diff(self, adata: AnnData):
        for n_states in range(2, 5):
            e = cr.tl.terminal_states(
                adata,
                estimator=cr.tl.estimators.GPCCA,
                cluster_key="clusters",
                method="brandts",
                show_plots=True,
                n_states=n_states,
                fit_kwargs={"n_cells": 5},
                return_estimator=True,
            )
            assert e.absorption_probabilities is None

            cr.tl.lineages(adata, backward=False)
            cr.pl.lineages(adata)

            _assert_has_all_keys(adata, Direction.FORWARD)
            probs = adata.obsm[str(AbsProbKey.FORWARD)]

            assert probs.shape[1] == n_states
            for name in probs.names:
                assert "Lineage" not in name, probs.names

    def test_multiple_read_write_same(self, adata: AnnData):
        e = cr.tl.terminal_states(
            adata,
            estimator=cr.tl.estimators.GPCCA,
            cluster_key="clusters",
            method="brandts",
            show_plots=True,
            n_states=2,
            fit_kwargs={"n_cells": 5},
            return_estimator=True,
        )
        cr.tl.lineages(adata, backward=False)
        cr.pl.lineages(adata)

        e.set_terminal_states({"foo": adata.obs_names[:20]})

        e = cr.tl.estimators.GPCCA(adata, read_from_adata=True, obsp_key="T_fwd")
        assert e.macrostates is None
        assert e.absorption_probabilities is None


class TestLowLevelPipeline:
    def test_fwd_pipelne_cflare(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()
        final_kernel = 0.8 * vk + 0.2 * ck

        estimator_fwd = cr.tl.estimators.CFLARE(final_kernel)

        estimator_fwd.compute_partition()

        estimator_fwd.compute_eigendecomposition()
        estimator_fwd.plot_spectrum()
        estimator_fwd.plot_spectrum(real_only=True)
        estimator_fwd.plot_eigendecomposition()
        estimator_fwd.plot_eigendecomposition(left=False)

        estimator_fwd.compute_terminal_states(use=1)
        estimator_fwd.plot_terminal_states()

        estimator_fwd.compute_absorption_probabilities()
        estimator_fwd.plot_absorption_probabilities()

        estimator_fwd.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

        _assert_has_all_keys(adata, Direction.FORWARD)

    def test_bwd_pipelne_cflare(self, adata: AnnData):
        vk = VelocityKernel(adata, backward=True).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata, backward=True).compute_transition_matrix()
        final_kernel = 0.8 * vk + 0.2 * ck

        estimator_bwd = cr.tl.estimators.CFLARE(final_kernel)

        estimator_bwd.compute_partition()

        estimator_bwd.compute_eigendecomposition()
        estimator_bwd.plot_spectrum()
        estimator_bwd.plot_spectrum(real_only=True)
        estimator_bwd.plot_eigendecomposition()
        estimator_bwd.plot_eigendecomposition(left=False)

        estimator_bwd.compute_terminal_states(use=1)
        estimator_bwd.plot_terminal_states()

        estimator_bwd.compute_absorption_probabilities()
        estimator_bwd.plot_absorption_probabilities()

        estimator_bwd.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

        _assert_has_all_keys(adata, Direction.BACKWARD)

    def test_fwd_pipelne_gpcca(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()
        final_kernel = 0.8 * vk + 0.2 * ck

        estimator_fwd = cr.tl.estimators.GPCCA(final_kernel)

        estimator_fwd.compute_partition()

        estimator_fwd.compute_eigendecomposition()
        estimator_fwd.plot_spectrum()
        estimator_fwd.plot_spectrum(real_only=True)

        estimator_fwd.compute_schur(5, method="brandts")
        estimator_fwd.plot_schur()

        estimator_fwd.compute_macrostates(3, n_cells=10)
        estimator_fwd.plot_macrostates()
        estimator_fwd.plot_coarse_T(show_initial_dist=True, show_stationary_dist=True)
        estimator_fwd.plot_schur_matrix()

        # select all states
        estimator_fwd.set_terminal_states_from_macrostates(n_cells=10)
        estimator_fwd.plot_terminal_states()

        estimator_fwd.compute_absorption_probabilities()
        estimator_fwd.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

        _assert_has_all_keys(adata, Direction.FORWARD)

        # select a subset of states
        estimator_fwd.set_terminal_states_from_macrostates(
            n_cells=16,
            names=estimator_fwd._get(P.MACRO).cat.categories[:2],
        )
        estimator_fwd.plot_terminal_states()

        estimator_fwd.compute_absorption_probabilities()
        estimator_fwd.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

        _assert_has_all_keys(adata, Direction.FORWARD)

    def test_bwd_pipelne_gpcca(self, adata: AnnData):
        vk = VelocityKernel(adata, backward=True).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(adata, backward=True).compute_transition_matrix()
        final_kernel = 0.8 * vk + 0.2 * ck

        estimator_bwd = cr.tl.estimators.GPCCA(final_kernel)

        estimator_bwd.compute_eigendecomposition()
        estimator_bwd.plot_spectrum()
        estimator_bwd.plot_spectrum(real_only=True)

        estimator_bwd.compute_schur(5, method="brandts")
        estimator_bwd.plot_schur()

        estimator_bwd.compute_macrostates(3, n_cells=16)
        estimator_bwd.plot_macrostates()
        estimator_bwd.plot_coarse_T(show_initial_dist=True, show_stationary_dist=True)
        estimator_bwd.plot_schur_matrix()

        # select all cells
        estimator_bwd.set_terminal_states_from_macrostates(n_cells=16)
        estimator_bwd.plot_terminal_states()

        estimator_bwd.compute_absorption_probabilities()
        estimator_bwd.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

        _assert_has_all_keys(adata, Direction.BACKWARD)

        # select a subset of states
        estimator_bwd.set_terminal_states_from_macrostates(
            n_cells=16,
            names=estimator_bwd._get(P.MACRO).cat.categories[:2],
        )
        estimator_bwd.plot_terminal_states()

        estimator_bwd.compute_absorption_probabilities()
        estimator_bwd.compute_lineage_drivers(cluster_key="clusters", use_raw=False)

        _assert_has_all_keys(adata, Direction.BACKWARD)



"""Matrix decomposition module."""
from abc import ABC
from typing import Any, Tuple, Union, Mapping, Optional
from pathlib import Path

from pygpcca import GPCCA as _GPCCA
from pygpcca._sorted_schur import _check_conj_split

from cellrank import logging as logg
from cellrank.ul._docs import d, inject_docs
from cellrank.tl._utils import save_fig, _eigengap
from cellrank.tl.estimators._utils import Metadata, _delegate
from cellrank.tl.estimators._property import Property, KernelHolder, VectorPlottable
from cellrank.tl.estimators._constants import A, F, P

import numpy as np
from scipy.sparse import issparse
from scipy.sparse.linalg import eigs

import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, FormatStrFormatter
from mpl_toolkits.axes_grid1 import make_axes_locatable

EPS = np.finfo(np.float64).eps


class Decomposable(KernelHolder, Property, ABC):
    """Helper class exposes writing the eigendecomposition to :class:`anndata.AnnData` object."""

    def _write_eig_to_adata(
        self, eig: Mapping[str, Any], start=None, extra_msg: Optional[str] = None
    ):
        setattr(self, A.EIG.s, eig)
        self.adata.uns[f"eig_{self._direction}"] = eig

        msg = f"Adding `adata.uns['eig_{self._direction}']`\n       `.{P.EIG}`"
        if extra_msg is None:
            extra_msg = "\n    Finish"

        msg += extra_msg

        logg.info(msg, time=start)


class Eigen(VectorPlottable, Decomposable):
    """Class computing the eigendecomposition."""

    __prop_metadata__ = [
        Metadata(attr=A.EIG, prop=P.EIG, dtype=Mapping[str, Any], compute_fmt=F.NO_FUNC)
    ]

    @d.dedent
    @inject_docs(prop=P.EIG)
    def compute_eigendecomposition(
        self,
        k: int = 20,
        which: str = "LR",
        alpha: float = 1,
        only_evals: bool = False,
        ncv: Optional[int] = None,
    ) -> None:
        """
        Compute eigendecomposition of transition matrix.

        Uses a sparse implementation, if possible, and only computes the top :math:`k` eigenvectors
        to speed up the computation. Computes both left and right eigenvectors.

        Parameters
        ----------
        k
            Number of eigenvalues/vectors to compute.
        %(eigen)s
        only_evals
            Compute only eigenvalues.
        ncv
            Number of Lanczos vectors generated.

        Returns
        -------
        None
            Nothing, but updates the following field:

                - :attr:`{prop}`
        """

        def get_top_k_evals():
            return D[np.flip(np.argsort(D.real))][:k]

        start = logg.info("Computing eigendecomposition of the transition matrix")

        if self.issparse:
            logg.debug(f"Computing top `{k}` eigenvalues for sparse matrix")
            D, V_l = eigs(self.transition_matrix.T, k=k, which=which, ncv=ncv)
            if only_evals:
                self._write_eig_to_adata(
                    {
                        "D": get_top_k_evals(),
                        "eigengap": _eigengap(get_top_k_evals().real, alpha),
                        "params": {"which": which, "k": k, "alpha": alpha},
                    }
                )
                return
            _, V_r = eigs(self.transition_matrix, k=k, which=which, ncv=ncv)
        else:
            logg.warning(
                "This transition matrix is not sparse, computing full eigendecomposition"
            )
            D, V_l = np.linalg.eig(self.transition_matrix.T)
            if only_evals:
                self._write_eig_to_adata(
                    {
                        "D": get_top_k_evals(),
                        "eigengap": _eigengap(D.real, alpha),
                        "params": {"which": which, "k": k, "alpha": alpha},
                    }
                )
                return
            _, V_r = np.linalg.eig(self.transition_matrix)

        # Sort the eigenvalues and eigenvectors and take the real part
        logg.debug("Sorting eigenvalues by their real part")
        p = np.flip(np.argsort(D.real))
        D, V_l, V_r = D[p], V_l[:, p], V_r[:, p]
        e_gap = _eigengap(D.real, alpha)

        pi = np.abs(V_l[:, 0].real)
        pi /= np.sum(pi)

        self._write_eig_to_adata(
            {
                "D": D,
                "stationary_dist": pi,
                "V_l": V_l,
                "V_r": V_r,
                "eigengap": e_gap,
                "params": {"which": which, "k": k, "alpha": alpha},
            },
            start=start,
        )

    @d.dedent
    def plot_eigendecomposition(self, left: bool = False, *args, **kwargs):
        """
        Plot eigenvectors in an embedding.

        Parameters
        ----------
        left
            Whether to plot left or right eigenvectors.
        %(plot_vectors.parameters)s

        Returns
        -------
        %(plot_vectors.returns)s
        """

        eig = getattr(self, P.EIG.s, None)

        if eig is None:
            self._plot_vectors(None, P.EIG.s)

        side = "left" if left else "right"
        D, V = (
            eig["D"],
            eig.get(f"V_{side[0]}", None),
        )
        if V is None:
            raise RuntimeError(
                "Compute eigendecomposition first as `.compute_eigendecomposition(..., only_evals=False)`."
            )

        # if irreducible, first rigth e-vec should be const.
        if side == "right":
            # quick check for irreducibility:
            if np.sum(np.isclose(D, 1, rtol=1e2 * EPS, atol=1e2 * EPS)) == 1:
                V[:, 0] = 1.0

        self._plot_vectors(
            V,
            P.EIG.s,
            *args,
            D=D,
            **kwargs,
        )

    @d.dedent
    def plot_spectrum(
        self,
        n: Optional[int] = None,
        real_only: bool = False,
        show_eigengap: bool = True,
        show_all_xticks: bool = True,
        legend_loc: Optional[str] = None,
        title: Optional[str] = None,
        figsize: Optional[Tuple[float, float]] = (5, 5),
        dpi: int = 100,
        save: Optional[Union[str, Path]] = None,
        marker: str = ".",
        **kwargs,
    ) -> None:
        """
        Plot the top eigenvalues in real or complex plane.

        Parameters
        ----------
        n
            Number of eigenvalues to show. If `None`, show all that have been computed.
        real_only
            Whether to plot only the real part of the spectrum.
        show_eigengap
            When `real_only=True`, this determines whether to show the inferred eigengap as
            a dotted line.
        show_all_xticks
            When `real_only=True`, this determines whether to show the indices of all eigenvalues
            on the x-axis.
        legend_loc
            Location parameter for the legend.
        title
            Title of the figure.
        %(plotting)s
        marker
            Marker symbol used, valid options can be found in :mod:`matplotlib.markers`.
        kwargs
            Keyword arguments for :func:`matplotlib.pyplot.scatter`.

        Returns
        -------
        %(just_plots)s
        """

        eig = getattr(self, P.EIG.s, None)
        if eig is None:
            raise RuntimeError(
                f"Compute `.{P.EIG}` first as `.{F.COMPUTE.fmt(P.EIG)}()`."
            )
        if n is None:
            n = len(eig["D"])
        elif n <= 0:
            raise ValueError(f"Expected `n` to be > 0, found `{n}`.")

        if real_only:
            fig = self._plot_real_spectrum(
                n,
                show_eigengap=show_eigengap,
                show_all_xticks=show_all_xticks,
                dpi=dpi,
                figsize=figsize,
                legend_loc=legend_loc,
                title=title,
                marker=marker,
                **kwargs,
            )
        else:
            fig = self._plot_complex_spectrum(
                n,
                dpi=dpi,
                figsize=figsize,
                legend_loc=legend_loc,
                title=title,
                marker=marker,
                **kwargs,
            )

        if save:
            save_fig(fig, save)

    def _plot_complex_spectrum(
        self,
        n: int,
        dpi: int = 100,
        figsize: Optional[Tuple[float, float]] = (None, None),
        legend_loc: Optional[str] = None,
        title: Optional[str] = None,
        marker: str = ".",
        **kwargs,
    ):
        # define a function to make the data limits rectangular
        def adapt_range(min_, max_, range_):
            return (
                min_ + (max_ - min_) / 2 - range_ / 2,
                min_ + (max_ - min_) / 2 + range_ / 2,
            )

        eig = getattr(self, P.EIG.s)
        D, params = eig["D"][:n], eig["params"]

        # create fiture and axes
        fig, ax = plt.subplots(nrows=1, ncols=1, dpi=dpi, figsize=figsize)

        # get the original data ranges
        lam_x, lam_y = D.real, D.imag
        x_min, x_max = np.min(lam_x), np.max(lam_x)
        y_min, y_max = np.min(lam_y), np.max(lam_y)
        x_range, y_range = x_max - x_min, y_max - y_min
        final_range = np.max([x_range, y_range]) + 0.05

        x_min_, x_max_ = adapt_range(x_min, x_max, final_range)
        y_min_, y_max_ = adapt_range(y_min, y_max, final_range)

        # plot the data and the unit circle
        ax.scatter(D.real, D.imag, marker=marker, label="eigenvalue", **kwargs)
        t = np.linspace(0, 2 * np.pi, 500)
        x_circle, y_circle = np.sin(t), np.cos(t)
        ax.plot(x_circle, y_circle, "k-", label="unit circle")

        # set labels, ranges and legend
        ax.set_xlabel(r"Re($\lambda$)")
        ax.set_xlim(x_min_, x_max_)

        ax.set_ylabel(r"Im($\lambda$)")
        ax.set_ylim(y_min_, y_max_)

        key = "real part" if params["which"] == "LR" else "magnitude"
        if title is None:
            title = f"top {n} eigenvalues according to their {key}"

        ax.set_title(title)
        if legend_loc not in (None, "none"):
            ax.legend(loc=legend_loc)

        return fig

    def _plot_real_spectrum(
        self,
        n: int,
        show_eigengap: bool = True,
        show_all_xticks: bool = True,
        dpi: int = 100,
        figsize: Optional[Tuple[float, float]] = None,
        legend_loc: Optional[str] = None,
        title: Optional[str] = None,
        marker: str = ".",
        **kwargs,
    ):
        eig = getattr(self, P.EIG.s)
        D, params = eig["D"][:n], eig["params"]

        D_real, D_imag = D.real, D.imag
        ixs = np.arange(len(D))
        mask = D_imag == 0

        # plot the top eigenvalues
        fig, ax = plt.subplots(nrows=1, ncols=1, dpi=dpi, figsize=figsize)
        if np.any(mask):
            ax.scatter(
                ixs[mask],
                D_real[mask],
                marker=marker,
                label="real eigenvalue",
                **kwargs,
            )
        if np.any(~mask):
            ax.scatter(
                ixs[~mask],
                D_real[~mask],
                marker=marker,
                label="complex eigenvalue",
                **kwargs,
            )

        # add dashed line for the eigengap, ticks, labels, title and legend
        if show_eigengap and eig["eigengap"] < n:
            ax.axvline(eig["eigengap"], label="eigengap", ls="--", lw=2, c="k")

        ax.set_xlabel("index")
        if show_all_xticks:
            ax.set_xticks(np.arange(len(D)))
        else:
            ax.xaxis.set_major_locator(MultipleLocator(2.0))
            ax.xaxis.set_major_formatter(FormatStrFormatter("%d"))

        ax.set_ylabel(r"Re($\lambda_i$)")

        key = "real part" if params["which"] == "LR" else "magnitude"
        if title is None:
            title = f"real part of top {n} eigenvalues according to their {key}"

        ax.set_title(title)
        if legend_loc not in (None, "none"):
            ax.legend(loc=legend_loc)

        return fig


class Schur(VectorPlottable, Decomposable):
    """Class computing the Schur decomposition."""

    __prop_metadata__ = [
        Metadata(
            attr=A.SCHUR,
            prop=P.SCHUR,
            dtype=np.ndarray,
            compute_fmt=F.NO_FUNC,
            doc="Schur vectors.",
        ),
        Metadata(attr=A.SCHUR_MAT, prop=P.SCHUR_MAT, dtype=np.ndarray),
        Metadata(attr=A.EIG, prop=P.EIG, dtype=Mapping[str, Any]),
        Metadata(attr="_invalid_n_states", prop=P.NO_PROPERTY, dtype=np.ndarray),
        Metadata(attr="_gpcca", prop=P.NO_PROPERTY),
    ]

    @d.dedent
    @inject_docs(schur_vectors=P.SCHUR, schur_matrix=P.SCHUR_MAT, eigendec=P.EIG)
    def compute_schur(
        self,
        n_components: int = 10,
        initial_distribution: Optional[np.ndarray] = None,
        method: str = "krylov",
        which: str = "LR",
        alpha: float = 1,
    ):
        """
        Compute the Schur decomposition.

        Parameters
        ----------
        n_components
            Number of vectors to compute.
        initial_distribution
            Input probability distribution over all cells. If `None`, uniform is chosen.
        method
            Method for calculating the Schur vectors. Valid options are: `'krylov'` or `'brandts'`.
            For benefits of each method, see :class:`pygpcca.GPCCA`.

            The former is an iterative procedure that computes a partial, sorted Schur decomposition for large, sparse
            matrices whereas the latter computes a full sorted Schur decomposition of a dense matrix.
        %(eigen)s

        Returns
        -------
        None
            Nothing, but updates the following fields:

                - :attr:`{schur_vectors}`
                - :attr:`{schur_matrix}`
                - :attr:`{eigendec}`
        """
        if n_components < 2:
            raise ValueError(
                f"Number of components must be `>=2`, found `{n_components}`."
            )

        if method not in ("brandts", "krylov"):
            raise ValueError(
                f"Invalid method `{method!r}`. Valid options are: 'brandts', 'krylov'."
            )

        try:
            import petsc4py  # noqa
            import slepc4py  # noqa
        except ImportError:
            method = "brandts"
            logg.warning(
                f"Unable to import `petsc4py` or `slepc4py`. Using `method={method!r}`"
            )

        if method == "brandts" and issparse(self.transition_matrix):
            logg.warning("For `method='brandts'`, dense matrix is required. Densifying")
            tmat = self.transition_matrix.A
        else:
            tmat = self.transition_matrix

        self._gpcca = _GPCCA(tmat, eta=initial_distribution, z=which, method=method)
        start = logg.info("Computing Schur decomposition")

        try:
            self._gpcca._do_schur_helper(n_components)
        except ValueError:
            logg.warning(
                f"Using `{n_components}` components would split a block of complex conjugates. "
                f"Increasing `n_components` to `{n_components + 1}`"
            )
            self._gpcca._do_schur_helper(n_components + 1)

        # make it available for plotting
        setattr(self, A.SCHUR.s, self._gpcca._p_X)
        setattr(self, A.SCHUR_MAT.s, self._gpcca._p_R)

        self._invalid_n_states = np.array(
            [
                i
                for i in range(2, len(self._gpcca._p_eigenvalues))
                if _check_conj_split(self._gpcca._p_eigenvalues[:i])
            ]
        )
        if len(self._invalid_n_states):
            logg.info(
                f"When computing macrostates, choose a number of states NOT in `{list(self._invalid_n_states)}`"
            )

        self._write_eig_to_adata(
            {
                "D": self._gpcca._p_eigenvalues,
                "eigengap": _eigengap(self._gpcca._p_eigenvalues, alpha),
                "params": {
                    "which": which,
                    "k": len(self._gpcca._p_eigenvalues),
                    "alpha": alpha,
                },
            },
            start=start,
            extra_msg=f"\n       `.{P.SCHUR}`\n       `.{P.SCHUR_MAT}`\n    Finish",
        )

    plot_schur = _delegate(prop_name=P.SCHUR.s)(VectorPlottable._plot_vectors)

    @d.dedent
    def plot_schur_matrix(
        self,
        title: Optional[str] = "schur matrix",
        cmap: str = "viridis",
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[float] = 80,
        save: Optional[Union[str, Path]] = None,
        **kwargs,
    ):
        """
        Plot the Schur matrix.

        Parameters
        ----------
        title
            Title of the figure.
        cmap
            Colormap to use.
        %(plotting)s
        kwargs
            Keyword arguments for :func:`seaborn.heatmap`.

        Returns
        -------
        %(just_plots)s
        """

        from seaborn import heatmap

        schur_matrix = getattr(self, P.SCHUR_MAT.s, None)

        if schur_matrix is None:
            raise RuntimeError(
                f"Compute Schur matrix first as `.{F.COMPUTE.fmt(P.SCHUR)}()`."
            )

        fig, ax = plt.subplots(
            figsize=schur_matrix.shape if figsize is None else figsize, dpi=dpi
        )

        divider = make_axes_locatable(ax)  # square=True make the colorbar a bit bigger
        cbar_ax = divider.append_axes("right", size="2%", pad=0.1)

        mask = np.zeros_like(schur_matrix, dtype=np.bool_)
        mask[np.tril_indices_from(mask, k=-1)] = True
        mask[~np.isclose(schur_matrix, 0.0)] = False

        vmin, vmax = (
            np.min(schur_matrix[~mask]),
            np.max(schur_matrix[~mask]),
        )

        kwargs["fmt"] = kwargs.get("fmt", "0.2f")
        heatmap(
            schur_matrix,
            cmap=cmap,
            square=True,
            annot=True,
            vmin=vmin,
            vmax=vmax,
            cbar_ax=cbar_ax,
            cbar_kws={"ticks": np.linspace(vmin, vmax, 10)},
            mask=mask,
            xticklabels=[],
            yticklabels=[],
            ax=ax,
            **kwargs,
        )

        ax.set_title(title)

        if save is not None:
            save_fig(fig, path=save)



"""
Compute macrostates
-------------------

This example shows how to compute and plot the macrostates.

For the computation of macrostates, we adapted the popular Generalized Perron Cluster Cluster Analysis :cite:`reuter:18`
:cite:`reuter:19` method to the single cell context. We provide a scalable implementation which can decompose datasets
of 100k+ cells into their dominant dynamical macrostates in just a few minutes. GPCCA relies on the real Schur
decomposition to handle non-symmetric transition matrices as they arise from RNA velocity information.
"""

import cellrank as cr

adata = cr.datasets.pancreas_preprocessed("../example.h5ad")
adata

# %%
# First, we prepare the kernel using the high-level pipeline and the :class:`cellrank.tl.estimators.GPCCA` estimator.
k = cr.tl.transition_matrix(
    adata, weight_connectivities=0.2, softmax_scale=4, show_progress_bar=False
)
g = cr.tl.estimators.GPCCA(k)

# %%
# First, we need to compute the Schur vectors. By default, the first 10 vectors are computed.
g.compute_schur(n_components=4)

# %%
# We can now compute the macrostates of the Markov chain. The first important parameter is the ``cluster_key``,
# which tries to associate the names of macrostates with the cluster labels.
# The second import parameter is ``n_cells``, which selects the top cells from each state based
# on the membership degree. By default, 30 cells are selected.
#
# Lastly, the parameter ``n_states`` can also be estimated by using either the `eigengap` or the `minChi` criterion from
# :cite:`reuter:18`.
g.compute_macrostates(n_states=3, cluster_key="clusters")

# %%
# After computing the macrostates, we can inspect them as follows. Below we show for each cell the membership
# degree of the macrostates.
g.macrostates_memberships

# %%
# To get the categorical observations, top ``n_cells`` for each macrostate, we can inspect the attribute below.
g.macrostates

# %%
# We can now plot the membership degree, as well as the categorical assignment.
g.plot_macrostates()
g.plot_macrostates(discrete=True)

# %%
# Both of these options are shown in the same plot, which is not always desirable.
# To change this, simply run the following.
g.plot_macrostates(same_plot=False)

# %%
# Method :meth:`cellrank.tl.estimators.GPCCA.compute_macrostates` also computes the coarse-grained transition
# matrix between the macrostates, see :ref:`sphx_glr_auto_examples_estimators_compute_coarse_T.py`.



"""Lineage class module."""

from types import FunctionType, MappingProxyType
from typing import List, Tuple, Union, Mapping, TypeVar, Callable, Iterable, Optional
from inspect import signature
from pathlib import Path
from functools import wraps
from itertools import combinations

from typing_extensions import Literal

from cellrank import logging as logg
from cellrank.ul._docs import d
from cellrank.tl._utils import save_fig, _convert_lineage_name, _unique_order_preserving
from cellrank.tl._colors import (
    _get_bg_fg_colors,
    _compute_mean_color,
    _create_categorical_colors,
)
from cellrank.tl._constants import Lin, ModeEnum

import numpy as np
import pandas as pd
from scipy.stats import entropy

import matplotlib.colors as c
import matplotlib.pyplot as plt

ColorLike = TypeVar("ColorLike")
_ERROR_NOT_ITERABLE = "Expected `{}` to be iterable, found type `{}`."
_ERROR_WRONG_SIZE = "Expected `{}` to be of size `{{}}`, found `{{}}`."

_HT_CELLS = 10  # head and tail cells to show
_HTML_REPR_THRESH = 100
_DUMMY_CELL = "<td style='text-align: right;'>...</td>"
_ORDER = "C"


class PrimingDegree(ModeEnum):
    """Priming degree method."""

    KL_DIVERGENCE = "kl_divergence"
    ENTROPY = "entropy"


def _at_least_2d(array: np.ndarray, dim: int):
    return np.expand_dims(array, dim) if array.ndim < 2 else array


def wrap(numpy_func: Callable) -> Callable:
    """
    Wrap an numpy function.

    Modifies functionality of some function (e.g. ignoring `.squeeze`, retaining dimensions).

    Parameters
    ----------
    numpy_func
        Function to be wrapped.

    Returns
    -------
    :class:`Callable`
        Wrapped function which takes a :class:`cellrank.tl.Lineage` and return :class:`cellrank.tl.Lineage`.
    """

    @wraps(numpy_func)
    def decorator(array, *args, **kwargs):
        if not isinstance(array, Lineage):
            raise TypeError(
                f"Expected array to be of type `Lineage`, found `{type(array).__name__}`."
            )
        if fname == "squeeze":
            return array
        if fname == "array_repr":
            return repr(array)

        if "axis" in kwargs:
            axis = kwargs["axis"]
        elif axis_ix < len(args):
            axis = args[axis_ix]
        else:
            axis = default_axis

        res = np.array(numpy_func(array.X, *args, **kwargs), copy=False)

        # handle expand_dim
        if res.ndim > 2:
            return array

        # handle reductions
        if not res.shape:
            return Lineage(np.array([[res]]), names=[fname], colors=["grey"])
        if res.shape == array.shape:
            return Lineage(res, names=array.names, colors=array.colors)

        res = np.expand_dims(res, axis)

        is_t = int(array._is_transposed)
        if is_t:
            res = res.T

        lin = None
        if res.shape[0] == array.shape[is_t]:
            lin = Lineage(
                res,
                names=[f"{fname} of {', '.join(array.names)}"],
                colors=["grey"],
            )

        if res.shape[1] == array.shape[1 - is_t]:
            lin = Lineage(
                res, names=[f"{fname} of {n}" for n in array.names], colors=array.colors
            )

        if lin is not None:
            return lin.T if is_t else lin

        raise RuntimeError(
            f"Unable to interpret result of function `{fname}` called "
            f"with args `{args}`, kwargs: `{kwargs}`."
        )

    params = signature(numpy_func).parameters
    if "axis" in params:
        axis_ix = list(params.keys()).index("axis") - 1
        default_axis = params["axis"].default
    else:
        axis_ix = 256
        default_axis = None
    assert (
        axis_ix >= 0
    ), f"Expected argument `'axis'` not to be first for function `{numpy_func.__name__}`."

    fname = numpy_func.__name__
    if fname == "amin":
        fname = "min"
    elif fname == "amax":
        fname = "max"

    return decorator


def _register_handled_functions():
    handled_fns = {}

    for attrname in dir(np):
        fn = getattr(np, attrname)
        if isinstance(fn, FunctionType):
            try:
                sig = signature(fn)
                if "axis" in sig.parameters.keys():
                    handled_fns[fn] = wrap(fn)
            except ValueError:
                pass

    handled_fns.pop(np.expand_dims, None)

    handled_fns[np.allclose] = wrap(np.allclose)
    handled_fns[np.array_repr] = wrap(np.array_repr)
    handled_fns[entropy] = wrap(entropy)  # qol change

    return handled_fns


_HANDLED_FUNCTIONS = _register_handled_functions()


class LineageMeta(type):
    """
    Metaclass for Lineaage.

    Registers functions which are handled by us and overloads common attributes, such as `.sum` with these functions.
    """

    __overloaded_functions__ = dict(  # noqa
        sum=np.sum,
        mean=np.mean,
        min=np.min,
        argmin=np.argmin,
        max=np.max,
        argmax=np.argmax,
        std=np.std,
        var=np.var,
        sort=np.sort,
        squeeze=np.squeeze,
        entropy=entropy,
    )

    def __new__(cls, clsname, superclasses, attributedict):  # noqa
        res = type.__new__(cls, clsname, superclasses, attributedict)
        for attrname, fn in LineageMeta.__overloaded_functions__.items():
            wrapped_fn = _HANDLED_FUNCTIONS.get(fn, None)
            if wrapped_fn:
                setattr(res, attrname, wrapped_fn)

        return res


class Lineage(np.ndarray, metaclass=LineageMeta):
    """
    Lightweight :class:`numpy.ndarray` wrapper that adds names and colors.

    Parameters
    ----------
    input_array
        Input array containing lineage probabilities, each lineage being stored in a column.
    names
        Names of the lineages.
    colors
        Colors of the lineages.
    """

    def __new__(
        cls,
        input_array: np.ndarray,
        *,
        names: Iterable[str],
        colors: Optional[Iterable[ColorLike]] = None,
    ) -> "Lineage":
        """Create and return a new object."""

        if not isinstance(input_array, np.ndarray):
            raise TypeError(
                f"Input array must be of type `numpy.ndarray`, found `{type(input_array).__name__!r}`."
            )

        if input_array.ndim == 1:
            input_array = np.expand_dims(input_array, -1)
        elif input_array.ndim > 2:
            raise ValueError(
                f"Input array must be 2-dimensional, found `{input_array.ndim}`."
            )

        if input_array.shape[0] == 0:
            raise ValueError("Expected number cells to be at least 1, found 0.")
        if input_array.shape[1] == 0:
            raise ValueError("Expected number of lineages to be at least 1, found 0.")

        obj = np.array(input_array, copy=True).view(cls)
        obj._n_lineages = obj.shape[1]
        obj._is_transposed = False
        obj.names = names  # these always create a copy, which is a good thing
        obj.colors = colors

        return obj

    def __array_finalize__(self, obj) -> None:
        if obj is None:
            return

        _names = getattr(obj, "_names", None)
        if _names is not None:
            self._names = _names
            self._n_lineages = len(_names)
            self._names_to_ixs = {n: i for i, n in enumerate(self.names)}
        else:
            self._names = None
            self._names_to_ixs = None
            self._n_lineages = getattr(
                obj, "_n_lineages", obj.shape[1] if obj.ndim == 2 else 0
            )

        self._colors = getattr(obj, "colors", None)
        self._is_transposed = getattr(obj, "_is_transposed", False)

    def _mixer(self, rows, mixtures):
        def update_entries(key):
            if key:
                res.append(self[rows, key].X.sum(1))
                # item = (key, rows) if self._is_transposed else (rows, key)
                # res.append(self[item].X.sum(int(not self._is_transposed)))
                names.append(" or ".join(self.names[key]))
                colors.append(_compute_mean_color(self.colors[key]))

        lin_kind = [_ for _ in mixtures if isinstance(_, Lin)]
        if len(lin_kind) > 1:
            raise ValueError(
                f"`Lin` enum is allowed only once in the expression, found `{lin_kind}`."
            )

        keys = [
            tuple(
                self._maybe_convert_names(
                    _convert_lineage_name(mixture), default=mixture
                )
            )
            if isinstance(mixture, str)
            else (mixture,)
            for mixture in mixtures
            if not (isinstance(mixture, Lin))
        ]
        keys = _unique_order_preserving(keys)

        # check the `keys` are unique
        overlap = [set(ks) for ks in keys]
        for c1, c2 in combinations(overlap, 2):
            overlap = c1 & c2
            if overlap:
                raise ValueError(
                    f"Found overlapping keys: `{self.names[list(overlap)]}`."
                )

        seen = set()
        names, colors, res = [], [], []
        for key in map(list, keys):
            seen.update(self.names[key])
            update_entries(key)

        if len(lin_kind) == 1:
            lin_kind = lin_kind[0]
            keys = [i for i, n in enumerate(self.names) if n not in seen]

            if lin_kind == Lin.OTHERS:
                for key in keys:
                    update_entries([key])
            elif lin_kind == Lin.REST:
                update_entries(keys)
                if keys:
                    names[-1] = str(lin_kind)
            else:
                raise ValueError(f"Invalid `Lin` enum `{lin_kind}`.")

        res = np.stack(res, axis=-1)
        return Lineage(res, names=names, colors=colors)

    def __getitem__(self, item) -> "Lineage":
        was_transposed = False
        if self._is_transposed:
            was_transposed = True
            self = self.T
            if isinstance(item, tuple):
                item = item[::-1]

        obj = self.__getitem(item)

        return obj.T if was_transposed else obj

    def __getitem(self, item):
        if isinstance(item, tuple):
            if len(item) > 2:
                raise ValueError(
                    f"Expected key to be of length `2`, found `{len(item)}`."
                )

            item = list(item)
            if item[0] is Ellipsis or item[0] is None:
                item[0] = range(self.shape[0])
            if len(item) == 2 and (item[1] is Ellipsis or item[1] is None):
                item[1] = range(self.shape[1])
            item = tuple(item)

        is_tuple_len_2 = (
            isinstance(item, tuple)
            and len(item) == 2
            and isinstance(
                item[0], (int, np.integer, range, slice, tuple, list, np.ndarray)
            )
        )
        if is_tuple_len_2:
            rows, col = item

            if isinstance(col, (int, np.integer, str)):
                col = [col]

            try:
                # slicing an array where row/col are like 2D indices
                if 1 < len(col) == len(rows) and len(rows) == self.shape[0]:
                    col = self._maybe_convert_names(col, make_unique=False)
                    return Lineage(
                        # never remove this expand_dims - it's critical
                        np.expand_dims(self.X[rows, col], axis=-1),
                        names=["mixture"],
                        colors=["#000000"],
                    )
            except TypeError:  # because of range
                pass

            if isinstance(col, (list, tuple, np.ndarray)):
                if any(
                    map(
                        lambda i: isinstance(i, Lin)
                        or (isinstance(i, str) and ("," in i or "or" in i)),
                        col,
                    )
                ):
                    return self._mixer(rows, col)
                col = self._maybe_convert_names(col)
                item = rows, col
        else:
            if isinstance(item, (int, np.integer, str)):
                item = [item]

            col = range(len(self.names))
            if isinstance(item, (tuple, list, np.ndarray)):
                if any(
                    map(
                        lambda i: isinstance(i, Lin)
                        or (isinstance(i, str) and ("," in i or "or" in i)),
                        item,
                    )
                ):
                    return self._mixer(slice(None, None, None), item)
                elif any(map(lambda i: isinstance(i, str), item)):
                    item = (slice(None, None, None), self._maybe_convert_names(item))
                    col = item[1]

        shape, row_order, col_order = None, None, None
        if (
            is_tuple_len_2
            and not isinstance(item[0], slice)
            and not isinstance(item[1], slice)
        ):
            item_0 = (
                np.array(item[0]) if not isinstance(item[0], np.ndarray) else item[0]
            )
            item_1 = (
                np.array(item[1]) if not isinstance(item[1], np.ndarray) else item[1]
            )
            item_0 = _at_least_2d(item_0, -1)
            item_1 = _at_least_2d(item_1, 0)

            if item_1.dtype == np.bool:
                if item_0.dtype != np.bool:
                    if not issubclass(item_0.dtype.type, np.integer):
                        raise TypeError(f"Invalid type `{item_0.dtype.type}`.")
                    row_order = (
                        item_0[:, 0]
                        if item_0.shape[0] == self.shape[0]
                        else np.argsort(np.argsort(item_0[:, 0]))
                    )
                    item_0 = _at_least_2d(np.isin(np.arange(self.shape[0]), item_0), -1)
                item = item_0 * item_1
                shape = np.max(np.sum(item, axis=0)), np.max(np.sum(item, axis=1))
                col = np.where(np.all(item_1, axis=0))[0]
            elif item_0.dtype == np.bool:
                if item_1.dtype != np.bool:
                    if not issubclass(item_1.dtype.type, np.integer):
                        raise TypeError(f"Invalid type `{item_1.dtype.type}`.")
                    col_order = (
                        item_1[0, :]
                        if item_1.shape[1] == self.shape[1]
                        else np.argsort(np.argsort(item_1[0, :]))
                    )
                    item_1 = _at_least_2d(np.isin(np.arange(self.shape[1]), item_1), 0)

                item = item_0 * item_1
                shape = np.max(np.sum(item, axis=0)), np.max(np.sum(item, axis=1))
                col = np.where(np.all(item_1, axis=0))[0]
            else:
                item = (item_0, item_1)

        obj = super().__getitem__(item)

        if shape is not None:  # keep the resulting shape
            obj = obj.reshape(shape)
        if row_order is not None:
            obj = obj[row_order, :]
        if col_order is not None:
            obj = obj[:, col_order]

        if isinstance(obj, Lineage):
            obj._names = np.atleast_1d(self.names[col])
            obj._colors = np.atleast_1d(self.colors[col])
            if col_order is not None:
                obj._names = obj._names[col_order]
                obj._colors = obj._colors[col_order]
            obj._names_to_ixs = {name: i for i, name in enumerate(obj._names)}

        return obj

    def __array_function__(self, func, types, args, kwargs):
        if func not in _HANDLED_FUNCTIONS:
            return NotImplemented
        # Note: this allows subclasses that don't override
        # __array_function__ to handle MyArray objects
        if not all(issubclass(t, type(self)) for t in types):
            return NotImplemented

        return _HANDLED_FUNCTIONS[func](*args, **kwargs)

    @property
    def names(self) -> np.ndarray:
        """Lineage names. Must be unique."""
        return self._names

    @names.setter
    def names(self, value: Iterable[str]):
        if not isinstance(value, Iterable):
            raise TypeError(_ERROR_NOT_ITERABLE.format("names", type(value).__name__))

        value = [v if isinstance(v, str) else str(v) for v in value]
        value = self._check_shape(value, _ERROR_WRONG_SIZE.format("names"))

        if len(set(value)) != len(value):
            raise ValueError(f"Not all lineage names are unique: `{value}`.")

        self._names = self._prepare_annotation(value)
        self._names_to_ixs = {name: ix for ix, name in enumerate(self.names)}

    @property
    def colors(self) -> np.ndarray:
        """Lineage colors."""
        return self._colors

    @colors.setter
    def colors(self, value: Optional[Iterable[ColorLike]]):
        if value is None:
            value = _create_categorical_colors(self._n_lineages)
        elif not isinstance(value, Iterable):
            raise TypeError(_ERROR_NOT_ITERABLE.format("colors", type(value).__name__))

        value = self._check_shape(value, _ERROR_WRONG_SIZE.format("colors"))
        self._colors = self._prepare_annotation(
            value,
            checker=c.is_color_like,
            transformer=c.to_hex,
            checker_msg="Value `{}` is not a valid color.",
        )

    def _maybe_convert_names(
        self,
        names: Iterable[Union[int, str, bool]],
        is_singleton: bool = False,
        default: Optional[Union[int, str]] = None,
        make_unique: bool = True,
    ) -> Union[int, List[int], List[bool]]:
        if all(map(lambda n: isinstance(n, (bool, np.bool_)), names)):
            return list(names)
        res = []
        for name in names:
            if isinstance(name, str):
                if name in self._names_to_ixs:
                    name = self._names_to_ixs[name]
                elif default is not None:
                    if isinstance(default, str):
                        if default not in self._names_to_ixs:
                            raise KeyError(
                                f"Invalid lineage name: `{name}`. "
                                f"Valid names are: `{list(self.names)}`."
                            )
                        name = self._names_to_ixs[default]
                    else:
                        name = default
                else:
                    raise KeyError(
                        f"Invalid lineage name `{name}`. "
                        f"Valid names are: `{list(self.names)}`."
                    )
            res.append(name)

        if make_unique:
            res = _unique_order_preserving(res)

        return res[0] if is_singleton else res

    def _check_shape(
        self, array: Iterable[Union[str, ColorLike]], msg: str
    ) -> List[Union[str, ColorLike]]:
        array = list(array)
        if len(array) != self._n_lineages:
            raise ValueError(msg.format(self._n_lineages, len(array)))

        return array

    def _prepare_annotation(
        self,
        array: List[str],
        checker: Optional[Callable] = None,
        transformer: Optional[Callable] = None,
        checker_msg: Optional[str] = None,
    ) -> np.ndarray:
        if checker:
            assert checker_msg, "Please provide a message when `checker` is not `None`."
            for v in array:
                if not checker(v):
                    raise ValueError(checker_msg.format(v))

        if transformer is not None:
            array = np.array([transformer(v) for v in array])

        return np.array(array)

    @property
    def X(self) -> np.ndarray:
        """Convert self to numpy array, losing names and colors."""
        return np.array(self, copy=False)

    @property
    def T(self):
        """Transpose of self."""
        obj = self.transpose()
        obj._is_transposed = not self._is_transposed
        return obj

    def view(self, dtype=None, type=None) -> "LineageView":
        """Return a view of self."""
        return LineageView(self)

    def __repr__(self) -> str:
        return f'{super().__repr__()[:-1]},\n  names([{", ".join(self.names)}]))'

    def __str__(self):
        return f'{super().__str__()}\n names=[{", ".join(self.names)}]'

    @property
    def _fmt(self) -> Callable:
        if np.issubdtype(self.dtype, np.float):
            return "{:.06f}".format
        return "{}".format

    def _repr_html_(self) -> str:
        def format_row(r):
            rng = (
                range(self.shape[1])
                if not self._is_transposed
                or (self._is_transposed and self.shape[1] <= _HTML_REPR_THRESH)
                else list(range(_HT_CELLS))
                + [...]
                + list(range(self.shape[1] - _HT_CELLS - 1, self.shape[1] - 1))
            )

            cells = "".join(
                f"<td style='text-align: right;'>" f"{self._fmt(self.X[r, c])}" f"</td>"
                if isinstance(c, int)
                else _DUMMY_CELL
                for c in rng
            )
            return f"<tr>{(names[r] if self._is_transposed else '') + cells}</tr>"

        def dummy_row():
            values = "".join(_DUMMY_CELL for _ in range(self.shape[1]))
            return f"<tr>{values}</tr>"

        if self.names is None or self.colors is None:
            raise RuntimeError(
                f"Name or colors are `None`. This can happen when running `array.view({type(self).__name__}`."
            )

        if self.ndim != 2:
            return repr(self)

        styles = [
            f"'background-color: {bg}; color: {fg}; text-align: center; word-wrap: break-word; max-width: 100px'"
            for bg, fg in map(_get_bg_fg_colors, self.colors)
        ]
        names = [f"<th style={style}>{n}</th>" for n, style in zip(self.names, styles)]
        header = f"<tr>{''.join(names)}</tr>"

        if self.shape[0] > _HTML_REPR_THRESH:
            body = "".join(format_row(i) for i in range(_HT_CELLS))
            body += dummy_row()
            body += "".join(
                format_row(i)
                for i in range(self.shape[0] - _HT_CELLS - 1, self.shape[0] - 1)
            )
        else:
            body = "".join(format_row(i) for i in range(self.shape[0]))

        cells = "cells" if self.shape[0] > 1 else "cell"
        lineages = "lineages" if self.shape[1] > 1 else "lineage"
        if self._is_transposed:
            cells, lineages = lineages, cells
        metadata = f"<p>{self.shape[0]} {cells} x {self.shape[1]} {lineages}</p>"

        if self._is_transposed:
            header = ""
        return (
            f"<div style='scoped' class='rendered_html'>"
            f"<table class='dataframe'>{header}{body}</table>{metadata}"
            f"</div>"
        )

    def __format__(self, format_spec):
        if self.shape == (1, 1):
            return format_spec.format(self.X[0, 0])
        if self.shape == (1,):
            return format_spec.format(self.X[0])
        return NotImplemented

    def __setstate__(self, state):
        *state, names, colors, is_t = state
        names = names[-1]
        colors = colors[-1]

        self._names = np.empty(names[1])
        self._colors = np.empty(colors[1])

        super().__setstate__(tuple(state))
        self._names.__setstate__(tuple(names))
        self._colors.__setstate__(tuple(colors))

        self._is_transposed = is_t
        self._n_lineages = len(self.names)
        self._names_to_ixs = {name: ix for ix, name in enumerate(self.names)}

    def __reduce__(self):
        res = list(super().__reduce__())

        names = self.names.__reduce__()
        colors = self.colors.__reduce__()
        res[-1] += (names, colors, self._is_transposed)

        return tuple(res)

    def copy(self, _="C") -> "Lineage":
        """Return a copy of itself."""
        obj = Lineage(
            self.T if self._is_transposed else self,
            names=np.array(self.names, copy=True, order=_ORDER),
            colors=np.array(self.colors, copy=True, order=_ORDER),
        )
        return obj.T if self._is_transposed else obj

    def __copy__(self):
        return self.copy()

    @d.get_full_description(base="lin_pd")
    @d.get_sections(base="lin_pd", sections=["Parameters", "Returns"])
    def priming_degree(
        self,
        method: Literal["kl_divergence", "entropy"] = "kl_divergence",
        early_cells: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        Compute the degree of lineage priming.

        This method computes how naive vs. committed each individual cell is.
        It returns a score where 0 stands for naive and 1 stands for committed.

        Parameters
        ----------
        method
            The method used to compute the degree of lineage priming. Valid options are:

                - `'kl_divergence'`: as in :cite:`velten:17`, computes KL-divergence between the fate probabilities of
                  a cell and the average fate probabilities. Computation of average fate probabilities can be restricted
                  to a set of user-defined ``early_cells``.
                - `'entropy'`: as in :cite:`setty:19`, computes entropy over a cell's fate probabilities.
        early_cells
            Cell ids or a mask marking early cells. If `None`, use all cells. Only used when ``method='kl_divergence'``.

        Returns
        -------
        The priming degree.
        """
        early_cells = (
            np.ones((len(self),), dtype=np.bool_)
            if early_cells is None
            else np.asarray(early_cells)
        )
        if not np.issubdtype(early_cells.dtype, np.bool_):
            early_cells = np.unique(early_cells)

        method = PrimingDegree(method)
        probs = self.X
        early_subset = probs[early_cells, :]
        if not len(early_subset):
            raise ValueError("No early cells have been specified.")

        if method == PrimingDegree.KL_DIVERGENCE:
            probs = np.nan_to_num(
                np.sum(probs * np.log2(probs / np.mean(early_subset, axis=0)), axis=1),
                nan=1.0,
                copy=False,
            )
        elif method == PrimingDegree.ENTROPY:
            probs = entropy(probs, axis=1)
            probs = np.max(probs) - probs
        else:
            raise NotImplementedError(f"Method `{method}` is not yet implemented")

        minn, maxx = np.min(probs), np.max(probs)
        return (probs - minn) / (maxx - minn)

    @d.dedent
    def plot_pie(
        self,
        reduction: Callable,
        title: Optional[str] = None,
        legend_loc: Optional[str] = "on data",
        legend_kwargs: Mapping = MappingProxyType({}),
        figsize: Optional[Tuple[float, float]] = None,
        dpi: Optional[float] = None,
        save: Optional[Union[Path, str]] = None,
        **kwargs,
    ) -> None:
        """
        Plot a pie chart visualizing aggregated lineage probabilities.

        Parameters
        ----------
        reduction
            Function that will be applied lineage-wise.
        title
            Title of the figure.
        legend_loc
            Location of the legend. If `None`, it is not shown.
        legend_kwargs
            Keyword arguments for :func:`matplotlib.axes.Axes.legend`.
        %(plotting)s

        Returns
        -------
        %(just_plots)s
        """

        if len(self.names) == 1:
            raise ValueError("Cannot plot pie chart for only 1 lineage.")

        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)

        if "autopct" not in kwargs:
            autopct_found = False
            autopct = (
                "{:.1f}%".format
            )  # we don't really care, we don't shot the pct, but the value
        else:
            autopct_found = True
            autopct = kwargs.pop("autopct")

        if title is None:
            title = reduction.__name__ if hasattr(reduction, "__name__") else None

        reduction = reduction(self, axis=int(self._is_transposed)).X.squeeze()
        reduction_norm = reduction / np.sum(reduction)

        wedges, texts, *autotexts = ax.pie(
            reduction_norm.squeeze(),
            labels=self.names if legend_loc == "on data" else None,
            autopct=autopct,
            wedgeprops={"edgecolor": "w"},
            colors=self.colors,
            **kwargs,
        )

        # if autopct is not None
        if len(autotexts):
            autotexts = autotexts[0]
            for name, at in zip(self.names, autotexts):
                ix = self._names_to_ixs[name]
                at.set_color(_get_bg_fg_colors(self.colors[ix])[1])
                if not autopct_found:
                    at.set_text(f"{reduction[ix]:.4f}")

        if legend_loc not in (None, "none", "on data"):
            ax.legend(
                wedges,
                self.names,
                title="lineages",
                loc=legend_loc,
                **legend_kwargs,
            )

        ax.set_title(title)
        ax.set_aspect("equal")

        fig.show()

        if save is not None:
            save_fig(fig, save)

    @d.dedent
    def reduce(
        self,
        *keys: str,
        mode: str = "dist",
        dist_measure: Literal[
            "cosine_sim", "wasserstein_dist", "kl_div", "js_div", "mutual_inf", "equal"
        ] = "mutual_info",
        normalize_weights: Literal["scale", "softmax"] = "softmax",
        softmax_scale: float = 1,
        return_weights: bool = False,
    ) -> Union["Lineage", Tuple["Lineage", Optional[pd.DataFrame]]]:
        """
        Subset states and normalize them so that they again sum to 1.

        Parameters
        ----------
        keys
            List of keys that define the states, to which this object will be reduced by projecting the values
            of the other states.
        mode
            Whether to use a distance measure to compute weights - `'dist'`, or just rescale - `'scale'`.
        dist_measure
            Used to quantify similarity between query and reference states. Valid options are:

                - `'cosine_sim'` - cosine similarity.
                - `'wasserstein_dist'` - Wasserstein distance.
                - `'kl_div'` - Kullback–Leibler divergence.
                - `'js_div'` - Jensen–Shannon divergence.
                - `'mutual_inf'` - mutual information.
                - `'equal'` - equally redistribute the mass among the rest.
        normalize_weights
            How to row-normalize the weights. Valid options are:

                - `'scale'` - divide by the sum.
                - `'softmax'`- use a softmax.
        softmax_scale
            Scaling factor in the softmax, used for normalizing the weights to sum to `1`.
        return_weights
            If `True`, a :class:`pandas.DataFrame` of the weights used for the projection is returned.

        Returns
        -------
        :class:`cellrank.tl.Lineage`, :class:`pandas.DataFrame`
            Lineage object, reduced to the %(initial_or_terminal)s states. If a reduction is not possible,
            returns just a copy of self.The weights used for the projection of shape
            ``(n_query, n_reference)``, if ``return_weights=True``.
        """

        if self._is_transposed:
            raise RuntimeError("This method works only on non-transposed lineages.")

        if not len(keys):
            raise ValueError("Unable to perform the reduction, no keys specified.")

        # check the lineage object
        if not np.allclose(np.sum(self, axis=1).X, 1.0):
            raise ValueError("Memberships do not sum to one row-wise.")

        if len(keys) == 1:
            tmp = self[:, keys]
            return Lineage(
                np.ones((self.shape[0], 1), dtype=self.dtype),
                names=tmp.names,
                colors=tmp.colors,
            )

        # check input parameters
        if return_weights and mode == "scale":
            logg.warning(
                "If `mode=='scale'`, no weights are computed. Returning `None`"
            )

        reference = self[:, keys]
        rest = [
            k for k in self.names if all(map(lambda rk: k not in rk, reference.names))
        ]
        if not len(rest):
            logg.warning(
                "Unable to perform reduction because all keys have been selected. Returning combined object only"
            )
            return (reference.copy(), None) if return_weights else reference.copy()

        query = self[:, rest]

        if mode == "dist":
            # compute a set of weights of shape (n_query x n_reference)
            if dist_measure == "cosine_sim":
                weights = _cosine_sim(reference.X, query.X)
            elif dist_measure == "wasserstein_dist":
                weights = _wasserstein_dist(reference.X, query.X)
            elif dist_measure == "kl_div":
                weights = _kl_div(reference.X, query.X)
            elif dist_measure == "js_div":
                weights = _js_div(reference.X, query.X)
            elif dist_measure == "mutual_info":
                weights = _mutual_info(reference.X, query.X)
            elif dist_measure == "equal":
                weights = _row_normalize(np.ones((query.shape[1], reference.shape[1])))
            else:
                raise ValueError(f"Invalid distance measure `{dist_measure!r}`.")

            # make some checks on the weights
            if weights.shape != (query.shape[1], reference.shape[1]):
                raise ValueError(
                    f"Expected weight matrix to be of shape `({query.shape[1]}, {reference.shape[1]})`, "
                    f"found `{weights.shape}`."
                )
            if not np.isfinite(weights).all():
                raise ValueError(
                    "Weights matrix contains elements that are not finite."
                )
            if (weights < 0).any():
                raise ValueError("Weights matrix contains negative elements.")

            if (weights == 0).any():
                logg.warning("Weights matrix contains exact zeros.")

            # normalize the weights to row-sum to one
            if normalize_weights == "scale":
                weights_n = _row_normalize(weights)
            elif normalize_weights == "softmax":
                weights_n = _softmax(_row_normalize(weights), softmax_scale)
            else:
                raise ValueError(
                    f"Normalization method `{normalize_weights!r}` not found. Valid options are: `'scale', 'softmax'`."
                )

            # check that the weights row-sum to one now
            if not np.allclose(weights_n.sum(1), 1.0):
                raise ValueError("Weights do not sum to 1 row-wise.")

            # use the weights to re-distribute probability mass form query to reference
            for i, w in enumerate(weights_n):
                reference += np.dot(query[:, i].X, w[None, :])

        elif mode == "scale":
            reference = _row_normalize(reference)
        else:
            raise ValueError(
                f"Invalid mode `{mode!r}`. Valid options are: `'dist', 'scale'`."
            )

        # check that the lineages row-sum to one now
        if not np.allclose(reference.sum(1), 1.0):
            raise ValueError("Reduced lineage rows do not sum to 1.")

        # potentially create a weights-df and return everything
        if return_weights:
            if mode == "dist":
                return (
                    reference,
                    pd.DataFrame(
                        data=weights_n, columns=reference.names, index=query.names
                    ),
                )
            return reference, None

        return reference


class LineageView(Lineage):
    """View of :class:`cellrank.tl.Lineage`."""

    def __new__(cls, lineage: Lineage) -> "LineageView":
        """Create a LineageView."""
        if not isinstance(lineage, Lineage):
            raise TypeError(
                f"Cannot create a `{cls.__name__}` of `{type(lineage).__name__}`."
            )

        view = np.array(lineage, copy=False).view(cls)
        view._owner = lineage
        view._names = lineage.names
        view._n_lineages = len(view.names)
        view._names_to_ixs = lineage._names_to_ixs
        view._colors = lineage.colors
        view._is_transposed = lineage._is_transposed

        return view

    @property
    def names(self) -> np.ndarray:
        """Lineage names."""
        return super().names

    @property
    def owner(self) -> Lineage:
        """Return the lineage associated with this view."""
        return self._owner

    @names.setter
    def names(self, _):
        raise RuntimeError(f"Unable to set names of `{type(self).__name__}`.")

    @property
    def colors(self) -> np.ndarray:
        """Lineage colors."""
        return super().colors

    @colors.setter
    def colors(self, _):
        raise RuntimeError(f"Unable to set colors of `{type(self).__name__}`.")

    def view(self, dtype=None, type=None) -> "LineageView":
        """Return self."""
        return self

    def copy(self, _="C") -> Lineage:
        """Return a copy of itself."""
        was_trasposed = False
        if self._is_transposed:
            self = self.T
            was_trasposed = True

        obj = Lineage(
            self,
            names=np.array(self.names, copy=True, order=_ORDER),
            colors=np.array(self.colors, copy=True, order=_ORDER),
        )
        return obj.T if was_trasposed else obj


def _remove_zero_rows(a: np.ndarray, b: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    if a.shape[0] != b.shape[0]:
        raise ValueError("Lineage objects have unequal cell numbers")

    bool_a = (a == 0).any(axis=1)
    bool_b = (b == 0).any(axis=1)
    mask = ~np.logical_or(bool_a, bool_b)

    logg.warning(
        f"Removed {a.shape[0] - np.sum(mask)} rows because they contained zeros"
    )

    return a[mask, :], b[mask, :]


def _softmax(X, beta: float = 1):
    return np.exp(X * beta) / np.expand_dims(np.sum(np.exp(X * beta), axis=1), -1)


def _row_normalize(X):
    if isinstance(X, Lineage):
        return X / X.sum(1)  # Lineage is shape-preserving
    return X / np.expand_dims(X.sum(1), -1)


def _col_normalize(X, norm_ord=2):
    from numpy.linalg import norm

    return X / norm(X, ord=norm_ord, axis=0)


def _cosine_sim(reference, query):
    # the cosine similarity is symmetric

    # normalize these to have 2-norm 1
    reference_n, query_n = _col_normalize(reference, 2), _col_normalize(query, 2)

    return (reference_n.T @ query_n).T


def _point_wise_distance(reference, query, distance):
    # utility function for all point-wise distances/divergences
    # take care of rows that contain zeros
    reference_no_zero, query_no_zero = _remove_zero_rows(reference, query)

    # normalize these to be valid probability distributions (column-wise)
    reference_n, query_n = (
        _col_normalize(reference_no_zero, 1),
        _col_normalize(query_no_zero, 1),
    )

    # loop over query and reference columns and compute pairwise wasserstein distances
    weights = np.zeros((query.shape[1], reference.shape[1]))
    for i, q_d in enumerate(query_n.T):
        for j, r_d in enumerate(reference_n.T):
            weights[i, j] = 1.0 / distance(q_d, r_d)

    return weights


def _wasserstein_dist(reference, query):
    # the wasserstein distance is symmetric
    from scipy.stats import wasserstein_distance

    return _point_wise_distance(reference, query, wasserstein_distance)


def _kl_div(reference, query):
    # the KL divergence is not symmetric
    from scipy.stats import entropy

    return _point_wise_distance(reference, query, entropy)


def _js_div(reference, query):
    # the js divergence is symmetric
    from scipy.spatial.distance import jensenshannon

    return _point_wise_distance(reference, query, jensenshannon)


def _mutual_info(reference, query):
    # mutual information is not symmetric. We don't need to normalise the vectors, it's invariant under scaling.
    from sklearn.feature_selection import mutual_info_regression

    weights = np.zeros((query.shape[1], reference.shape[1]))
    for i, target in enumerate(query.T):
        weights[i, :] = mutual_info_regression(reference, target)

    return weights



"""
Plot terminal states
--------------------

This example shows how to compute and plot the terminal states of the cell-state transition.

CellRank can be applied to any cell-state transition, be it differentiation, regeneration, reprogramming or other.
"""

import cellrank as cr

adata = cr.datasets.pancreas_preprocessed("../example.h5ad")
adata

# %%
# First, we compute the terminal states. By default, we're using the :class:`cellrank.tl.estimators.GPCCA` estimator.
# The parameter ``cluster_key`` tries to associate the names of the terminal states with cluster labels, whereas
# ``n_cells`` controls how many cells we take from each terminal state as categorical observation - this is only
# available to the above mentioned estimator. We can show some plots of interest by specifying ``show_plots=True``.
cr.tl.terminal_states(
    adata,
    cluster_key="clusters",
    n_cells=30,
    softmax_scale=4,
    n_states=3,
    show_progress_bar=False,
)

# %%
# We can now plot the terminal states. By default, when using :class:`cellrank.tl.estimators.GPCCA`, we plot continuous
# membership vectors to visualize individual cells associations with an terminal state.
cr.pl.terminal_states(adata)

# %%
# We can also plot membership vectors for different terminal states separately  state using ``same_plot=False``.
cr.pl.terminal_states(adata, same_plot=False)

# %%
# Lastly, we can discretize the assignment of cells to terminal states by showing the cells most
# likely to belong to the terminal state by specifying the ``discrete`` parameter.
cr.pl.terminal_states(adata, discrete=True)

# %%
# To see how to compute and plot the terminal states or the lineages, see
# :ref:`sphx_glr_auto_examples_plotting_plot_initial_states.py` or
# :ref:`sphx_glr_auto_examples_plotting_plot_lineages.py`, respectively.



"""Pseudotime kernel module."""
from copy import copy
from typing import Any, Union, Callable, Optional

from typing_extensions import Literal

from anndata import AnnData
from cellrank import logging as logg
from cellrank.ul._docs import d
from cellrank.tl._utils import _connected
from cellrank.tl.kernels import Kernel
from cellrank.tl._constants import ThresholdScheme
from cellrank.tl.kernels._base_kernel import _dtype
from cellrank.tl.kernels._pseudotime_schemes import (
    ThresholdSchemeABC,
    HardThresholdScheme,
    SoftThresholdScheme,
    CustomThresholdScheme,
)

import numpy as np


@d.dedent
class PseudotimeKernel(Kernel):
    """
    Kernel which computes directed transition probabilities based on a KNN graph and pseudotime.

    The KNN graph contains information about the (undirected) connectivities among cells, reflecting their similarity.
    Pseudotime can be used to either remove edges that point against the direction of increasing pseudotime
    :cite:`setty:19`, or to downweight them :cite:`stassen:21`.

    Parameters
    ----------
    %(adata)s
    %(backward)s
    time_key
        Key in :attr:`adata` ``.obs`` where the pseudotime is stored.
    %(cond_num)s
    kwargs
        Keyword arguments for :class:`cellrank.tl.kernels.Kernel`.
    """

    def __init__(
        self,
        adata: AnnData,
        backward: bool = False,
        time_key: str = "dpt_pseudotime",
        compute_cond_num: bool = False,
        check_connectivity: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            adata,
            backward=backward,
            time_key=time_key,
            compute_cond_num=compute_cond_num,
            check_connectivity=check_connectivity,
            **kwargs,
        )
        self._time_key = time_key

    def _read_from_adata(self, time_key: str, **kwargs: Any) -> None:
        super()._read_from_adata(**kwargs)

        if time_key not in self.adata.obs.keys():
            raise KeyError(f"Could not find time key in `adata.obs[{time_key!r}]`.")

        self._pseudotime = np.array(self.adata.obs[time_key]).astype(_dtype)
        if self.backward:
            self._pseudotime = np.max(self.pseudotime) - self.pseudotime

        if np.any(np.isnan(self._pseudotime)):
            raise ValueError("Encountered NaN values in pseudotime.")

    @d.dedent
    def compute_transition_matrix(
        self,
        threshold_scheme: Union[Literal["soft", "hard"], Callable] = "hard",
        frac_to_keep: float = 0.3,
        b: float = 10.0,
        nu: float = 0.5,
        check_irreducibility: bool = False,
        n_jobs: Optional[int] = None,
        backend: str = "loky",
        show_progress_bar: bool = True,
        **kwargs: Any,
    ) -> "PseudotimeKernel":
        """
        Compute transition matrix based on KNN graph and pseudotemporal ordering.

        Depending on the choice of the `thresholding_scheme`, this is based on ideas by either *Palantir*
        :cite:`setty:19` or *VIA* :cite:`stassen:21`.

        When using a `'hard'` thresholding scheme, this based on ideas by *Palantir* :cite:`setty:19` which removes some
        edges that point against the direction of increasing pseudotime. To avoid disconnecting the graph, it does not
        remove all edges that point against the direction of increasing pseudotime but keeps the ones that point to
        cells inside a close radius. This radius is chosen according to the local cell density.

        When using a `'soft'` thresholding scheme, this is based on ideas by *VIA* :cite:`stassen:21` which downweights
        edges that points against the direction of increasing pseudotime. Essentially, the further "behind" a query
        cell is in pseudotime with respect to the current reference cell, the more penalized will be its
        graph-connectivity.

        Parameters
        ----------
        frac_to_keep
            The `fract_to_keep` * n_neighbors closest neighbors (according to graph connectivities) are kept, no matter
            whether they lie in the pseudotemporal past or future. This is done to ensure that the graph remains
            connected. Only used when `threshold_scheme='hard'`.
        %(soft_scheme_kernel)s
        check_irreducibility
            Optional check for irreducibility of the final transition matrix.
        %(parallel)s
        kwargs
            Keyword arguments for ``threshold_scheme``.

        Returns
        -------
        :class:`cellrank.tl.kernels.PseudotimeKernel`
            Makes :attr:`transition_matrix` available.
        """
        start = logg.info(f"Computing transition matrix based on `{self._time_key}`")

        # get the connectivities and number of neighbors
        n_neighbors = (
            self.adata.uns.get("neighbors", {})
            .get("params", {})
            .get("n_neighbors", None)
        )
        if n_neighbors is None:
            logg.warning(
                "Could not find 'n_neighbors' in `adata.uns['neighbors']['params']`. Using an estimate"
            )
            n_neighbors = np.min(self._conn.sum(1))

        if isinstance(threshold_scheme, str):
            threshold_scheme = ThresholdScheme(threshold_scheme)
            if threshold_scheme == ThresholdScheme.SOFT:
                scheme = SoftThresholdScheme()
                kwargs["b"] = b
                kwargs["nu"] = nu
            elif threshold_scheme == ThresholdScheme.HARD:
                scheme = HardThresholdScheme()
                kwargs["frac_to_keep"], kwargs["n_neighs"] = frac_to_keep, n_neighbors
            else:
                raise NotImplementedError(
                    f"Threshold scheme `{threshold_scheme}` is not yet implemented."
                )
        elif isinstance(threshold_scheme, ThresholdSchemeABC):
            scheme = threshold_scheme
        elif callable(threshold_scheme):
            scheme = CustomThresholdScheme(threshold_scheme)
        else:
            raise TypeError(
                f"Expected `threshold_scheme` to be either a `str` or a `callable`, found `{type(threshold_scheme)}`."
            )

        # fmt: off
        if self._reuse_cache({"dnorm": False, "scheme": str(threshold_scheme), **kwargs}, time=start):
            return self
        # fmt: on

        biased_conn = scheme.bias_knn(
            self._conn,
            self.pseudotime,
            n_jobs=n_jobs,
            backend=backend,
            show_progress_bar=show_progress_bar,
            **kwargs,
        )

        # make sure the biased graph is still connected
        if not _connected(biased_conn):
            logg.warning("Biased KNN graph is disconnected")

        self._compute_transition_matrix(
            matrix=biased_conn,
            density_normalize=False,
            check_irreducibility=check_irreducibility,
        )
        logg.info("    Finish", time=start)

        return self

    @property
    def pseudotime(self) -> np.array:
        """Pseudotemporal ordering of cells."""
        return self._pseudotime

    def copy(self) -> "PseudotimeKernel":
        """Return a copy of self."""
        pk = PseudotimeKernel(
            self.adata, backward=self.backward, time_key=self._time_key
        )
        pk._pseudotime = copy(self.pseudotime)
        pk._params = copy(self._params)
        pk._cond_num = self.condition_number
        pk._transition_matrix = copy(self._transition_matrix)

        return pk

    def __invert__(self) -> "PseudotimeKernel":
        super().__invert__()
        self._pseudotime = np.max(self.pseudotime) - self.pseudotime
        return self



"""
Plot random walks
-----------------

This example shows how to plot random walks on a Markov chain in an embedding.
"""

import cellrank as cr

adata = cr.datasets.pancreas_preprocessed("../example.h5ad")
adata

# %%
# First, we create the kernel using the high-level function :func:`cellrank.tl.transition_matrix`, but any
# instance of :class:`cellrank.tl.kernels.Kernel` would do.
k = cr.tl.transition_matrix(
    adata, weight_connectivities=0.2, softmax_scale=4, show_progress_bar=False
)
k

# %%
# Next, we simulate and plot random walks. Below we simulate 100 random walks of length 100 where starting points
# are uniformly sampled from the `Ngn3 low EP` cluster.
k.plot_random_walks(
    100,
    start_ixs={"clusters": "Ngn3 low EP"},
    max_iter=100,
    show_progress_bar=False,
    ixs_legend_loc="best",
    seed=42,
)


# %%
# Apart from ``start_ixs``, we can also specify ``stop_ixs`` and ``successive_hits`` to control when random walks
# stop. Random walk is stopped if the maximum number of iterations is reached or when states in ``stop_ixs``
# is visited successively ``successive_hits`` times.
k.plot_random_walks(
    100,
    start_ixs={"clusters": "Ngn3 low EP"},
    stop_ixs={"clusters": ["Alpha", "Beta"]},
    max_iter=100,
    successive_hits=5,
    show_progress_bar=False,
    cmap="viridis",
    seed=42,
)

# %%
# Another way of visualizing the transition matrix is to project it onto an embedding,
# see :ref:`sphx_glr_auto_examples_kernels_plot_projection.py` for more information.



from copy import copy
from typing import Tuple, Callable, Optional

import pytest
from _helpers import (
    bias_knn,
    create_kernels,
    density_normalization,
    jax_not_installed_skip,
    random_transition_matrix,
)

import scanpy as sc
import cellrank as cr
from scanpy import Neighbors
from anndata import AnnData
from cellrank.tl._utils import _normalize
from cellrank.ul._utils import _get_neighs, _get_neighs_params
from cellrank.tl.kernels import (
    VelocityKernel,
    CytoTRACEKernel,
    PseudotimeKernel,
    PrecomputedKernel,
    ConnectivityKernel,
)
from cellrank.tl._constants import _transition
from cellrank.tl.kernels._base_kernel import (
    Kernel,
    Constant,
    KernelAdd,
    KernelMul,
    _dtype,
    _is_bin_mult,
)
from cellrank.tl.kernels._cytotrace_kernel import CytoTRACEAggregation, _ct

import numpy as np
from scipy.sparse import eye as speye
from scipy.sparse import isspmatrix_csr
from pandas.core.dtypes.common import is_bool_dtype, is_integer_dtype

_rtol = 1e-6


class CustomFunc(cr.tl.kernels.SimilaritySchemeABC):
    def __call__(
        self, v: np.ndarray, D: np.ndarray, softmax_scale: float = 1.0
    ) -> Tuple[np.ndarray, np.ndarray]:
        probs, logits = np.zeros((D.shape[0],), dtype=np.float64), np.zeros(
            (D.shape[0],), dtype=np.float64
        )
        probs[0] = 1.0

        return probs, logits


class CustomFuncHessian(CustomFunc):
    def hessian(
        self, v: np.ndarray, D: np.ndarray, _softmax_scale: float = 1.0
    ) -> np.ndarray:
        # should be either (n, g, g) or (n, g), will be (g, g)
        return np.zeros((D.shape[0], v.shape[0], v.shape[0]))


class CustomKernel(Kernel):
    def compute_transition_matrix(
        self, sparse: bool = False, dnorm: bool = False
    ) -> "KernelExpression":
        if sparse:
            tmat = speye(self.adata.n_obs, dtype=np.float32)
        else:
            tmat = np.eye(self.adata.n_obs, dtype=np.float32)

        self._compute_transition_matrix(tmat, density_normalize=dnorm)
        return self

    def copy(self) -> "KernelExpression":
        return copy(self)


class InvalidFuncProbs(cr.tl.kernels.SimilaritySchemeABC):
    def __call__(
        self, v: np.ndarray, D: np.ndarray, _softmax_scale: float = 1.0
    ) -> Tuple[np.ndarray, np.ndarray]:
        return np.ones((D.shape[0],), dtype=np.float64), np.zeros(
            (D.shape[0],), dtype=np.float64
        )


class InvalidFuncHessianShape(CustomFunc):
    def __call__(
        self, v: np.ndarray, D: np.ndarray, _softmax_scale: float = 1.0
    ) -> Tuple[np.ndarray, np.ndarray]:
        probs, logits = np.zeros((D.shape[0],), dtype=np.float64), np.zeros(
            (D.shape[0],), dtype=np.float64
        )
        probs[-1] = 1.0

        return probs, logits

    def hessian(
        self, v: np.ndarray, _D: np.ndarray, _softmax_scale: float = 1.0
    ) -> np.ndarray:
        # should be either (n, g, g) or (n, g), will be (g, g)
        return np.zeros((v.shape[0], v.shape[0]))


class TestInitializeKernel:
    def test_none_transition_matrix(self, adata: AnnData):
        vk = VelocityKernel(adata)
        ck = ConnectivityKernel(adata)
        pk = PseudotimeKernel(adata, time_key="latent_time")

        assert vk._transition_matrix is None
        assert ck._transition_matrix is None
        assert pk._transition_matrix is None

    def test_not_none_transition_matrix_compute(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()
        pk = PseudotimeKernel(adata, time_key="latent_time").compute_transition_matrix()

        assert vk.transition_matrix is not None
        assert ck.transition_matrix is not None
        assert pk.transition_matrix is not None

    def test_not_none_transition_matrix_accessor(self, adata: AnnData):
        vk = VelocityKernel(adata)
        ck = ConnectivityKernel(adata)
        pk = PseudotimeKernel(adata, time_key="latent_time")

        assert vk.transition_matrix is not None
        assert ck.transition_matrix is not None
        assert pk.transition_matrix is not None

    def test_adding_hidden_constants(self, adata: AnnData):
        k = VelocityKernel(adata) + ConnectivityKernel(adata)

        assert _is_bin_mult(k[0])
        assert isinstance(k[0], KernelMul)
        assert isinstance(k[0][0], Constant)
        assert isinstance(k[0][1], VelocityKernel)
        assert k[0][0].transition_matrix == 1.0

        assert _is_bin_mult(k[1])
        assert isinstance(k[1], KernelMul)
        assert isinstance(k[1][0], Constant)
        assert isinstance(k[1][1], ConnectivityKernel)
        assert k[1][0].transition_matrix == 1.0

    def test_length(self, adata: AnnData):
        k = VelocityKernel(adata) + ConnectivityKernel(adata)
        assert len(k) == 2

    def test_accessor_out_of_range(self, adata: AnnData):
        k = VelocityKernel(adata) + ConnectivityKernel(adata)

        with pytest.raises(IndexError):
            _ = k[2]

    def test_parent(self, adata: AnnData):
        vk = VelocityKernel(adata)
        ck = ConnectivityKernel(adata)
        k = vk + ck

        assert vk._parent._parent is k  # invisible constants
        assert ck._parent._parent is k
        assert k._parent is None

    def test_uninitialized_both(self, adata: AnnData):
        k = VelocityKernel(adata) + ConnectivityKernel(adata)

        with pytest.raises(RuntimeError):
            k.compute_transition_matrix()

    def test_uninitialized_one(self, adata: AnnData):
        k = (
            VelocityKernel(adata)
            + ConnectivityKernel(adata).compute_transition_matrix()
        )

        with pytest.raises(RuntimeError):
            k.compute_transition_matrix()

    def test_initialized(self, adata: AnnData):
        k = (
            VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
            + ConnectivityKernel(adata).compute_transition_matrix()
        )
        k.compute_transition_matrix()

        assert k.transition_matrix is not None

    def test_invalida_type(self, adata: AnnData):
        with pytest.raises(TypeError):
            _ = None * VelocityKernel(adata)

    def test_negative_constant(self, adata: AnnData):
        with pytest.raises(ValueError):
            _ = -1 * VelocityKernel(adata)

    def test_invalid_constant(self, adata: AnnData):
        with pytest.raises(TypeError):
            _ = Constant(None, None)

    def test_inversion(self, adata: AnnData):
        c = ConnectivityKernel(adata, backward=False)
        assert not c.backward

        nc = ~c
        assert nc.backward

    def test_inversion_inplace(self, adata: AnnData):
        c = ConnectivityKernel(adata, backward=False)

        assert not c.backward
        _ = ~c
        assert c.backward

    def test_inversion_propagation(self, adata: AnnData):
        c = ConnectivityKernel(adata, backward=False)
        v = VelocityKernel(adata, backward=False)
        k = ~(c + v)

        assert c.backward
        assert v.backward
        assert k.backward

    def test_inversion_recalculation(self, adata: AnnData):
        c = ConnectivityKernel(adata).compute_transition_matrix()
        z = ~(c + c)
        with pytest.raises(RuntimeError):
            z.compute_transition_matrix()

    def test_inversion_preservation_of_constants(self, adata: AnnData):
        c = ConnectivityKernel(adata).compute_transition_matrix()
        a = (3 * c + 1 * c).compute_transition_matrix()
        b = ~a
        c.compute_transition_matrix()

        assert a[0][0].transition_matrix == 3 / 4
        assert b[0][0].transition_matrix == 3 / 4
        assert a[1][0].transition_matrix == 1 / 4
        assert b[1][0].transition_matrix == 1 / 4

    def test_addition_simple(self, adata: AnnData):
        k = VelocityKernel(adata) + ConnectivityKernel(adata)

        assert isinstance(k, KernelAdd)

    def test_multiplication_simple(self, adata: AnnData):
        k = 10 * VelocityKernel(adata)
        c = _is_bin_mult(k)

        assert isinstance(c, Constant)
        assert c.transition_matrix == 10

    def test_multiplication_simple_normalization(self, adata: AnnData):
        k = 10 * VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        c = _is_bin_mult(k)

        assert c.transition_matrix == 10

    def test_constant(self, adata: AnnData):
        k = 9 * VelocityKernel(adata) + 1 * ConnectivityKernel(adata)
        c1, c2 = _is_bin_mult(k[0]), _is_bin_mult(k[1])

        assert c1.transition_matrix == 9
        assert c2.transition_matrix == 1

    def test_constant_normalize_2(self, adata: AnnData):
        k = (
            9 * VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
            + 1 * ConnectivityKernel(adata).compute_transition_matrix()
        )
        k.compute_transition_matrix()
        c1, c2 = _is_bin_mult(k[0]), _is_bin_mult(k[1])

        assert c1.transition_matrix == 9 / 10
        assert c2.transition_matrix == 1 / 10

    def test_constant_normalize_3(self, adata: AnnData):
        k = (
            VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
            + ConnectivityKernel(adata).compute_transition_matrix()
            + ConnectivityKernel(adata).compute_transition_matrix()
        )
        k.compute_transition_matrix()
        c1, c2, c3 = _is_bin_mult(k[0]), _is_bin_mult(k[1]), _is_bin_mult(k[2])

        assert c1.transition_matrix == 1 / 3
        assert c2.transition_matrix == 1 / 3
        assert c3.transition_matrix == 1 / 3

    def test_constant_wrong_parentheses(self, adata: AnnData):
        k = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4) + (
            ConnectivityKernel(adata).compute_transition_matrix()
            + ConnectivityKernel(adata).compute_transition_matrix()
        )
        k.compute_transition_matrix()
        c1, c2, c3 = _is_bin_mult(k[0]), _is_bin_mult(k[1]), _is_bin_mult(k[2])

        assert c1.transition_matrix == 1 / 3
        assert c2.transition_matrix == 1 / 3
        assert c3.transition_matrix == 1 / 3

    def test_constant_correct_parentheses(self, adata: AnnData):
        k = 1 * VelocityKernel(adata).compute_transition_matrix(softmax_scale=4) + 1 * (
            ConnectivityKernel(adata).compute_transition_matrix()
            + ConnectivityKernel(adata).compute_transition_matrix()
        )
        k.compute_transition_matrix()
        c1, c2, c3 = (
            _is_bin_mult(k[0]),
            _is_bin_mult(k[1][1][0]),
            _is_bin_mult(k[1][1][1]),
        )

        assert c1.transition_matrix == 1 / 2
        assert c2.transition_matrix == 1 / 2
        assert c3.transition_matrix == 1 / 2

    def test_adaptive_kernel_constants(self, adata: AnnData):
        ck1 = ConnectivityKernel(adata).compute_transition_matrix()
        ck1._mat_scaler = np.random.normal(size=(adata.n_obs, adata.n_obs))

        ck2 = ConnectivityKernel(adata).compute_transition_matrix()
        ck2._mat_scaler = np.random.normal(size=(adata.n_obs, adata.n_obs))

        k = (3 * ck1) ^ (1 * ck2)
        k.compute_transition_matrix()

        assert k[0][0]._value == 3 / 4
        assert k[1][0]._value == 1 / 4

    def test_adaptive_kernel_complex(self, adata: AnnData):
        ck1 = ConnectivityKernel(adata).compute_transition_matrix()
        ck1._mat_scaler = np.random.normal(size=(adata.n_obs, adata.n_obs))

        ck2 = ConnectivityKernel(adata).compute_transition_matrix()
        ck2._mat_scaler = np.random.normal(size=(adata.n_obs, adata.n_obs))

        ck3 = ConnectivityKernel(adata).compute_transition_matrix()
        ck3._mat_scaler = np.random.normal(size=(adata.n_obs, adata.n_obs))

        k = 4 * ((3 * ck1) ^ (1 * ck2)) + 2 * ck3
        k.compute_transition_matrix()

        assert k[0][0].transition_matrix == 4 / 6
        assert k[1][0].transition_matrix == 2 / 6
        assert k[0][1][0][0]._value == 3 / 4
        assert k[0][1][1][0]._value == 1 / 4

    def test_repr(self, adata: AnnData):
        rpr = repr(VelocityKernel(adata))

        assert rpr == f"<{VelocityKernel.__name__}>"

    def test_repr_inv(self, adata: AnnData):
        rpr = repr(~VelocityKernel(adata))

        assert rpr == f"~<{VelocityKernel.__name__}>"

    def test_repr_inv_comb(self, adata: AnnData):
        rpr = repr(~(VelocityKernel(adata) + ConnectivityKernel(adata)))

        assert (
            rpr
            == f"~((1 * <{VelocityKernel.__name__}>) + (1 * <{ConnectivityKernel.__name__}>))"
        )

    def test_str_repr_equiv_no_transition_matrix(self, adata: AnnData):
        vk = VelocityKernel(adata)
        string = str(vk)
        rpr = repr(vk)

        assert string == rpr
        assert string == f"<{VelocityKernel.__name__}>"

    def test_str(self, adata: AnnData):
        string = str(ConnectivityKernel(adata).compute_transition_matrix())

        assert (
            string == f"<{ConnectivityKernel.__name__}[dnorm=True, key=connectivities]>"
        )

    def test_str_inv(self, adata: AnnData):
        string = str(
            ConnectivityKernel(adata, backward=True).compute_transition_matrix()
        )

        assert (
            string
            == f"~<{ConnectivityKernel.__name__}[dnorm=True, key=connectivities]>"
        )

    def test_combination_correct_parameters(self, adata: AnnData):
        from cellrank.tl.kernels import CosineScheme

        k = VelocityKernel(adata).compute_transition_matrix(
            softmax_scale=4,
            seed=42,
            scheme="cosine",
        ) + (
            ConnectivityKernel(adata).compute_transition_matrix(density_normalize=False)
            + ConnectivityKernel(adata).compute_transition_matrix(
                density_normalize=True
            )
        )
        k.compute_transition_matrix()

        assert isinstance(k.params, dict)
        assert len(k.params) == 3
        assert {"dnorm": True, "key": "connectivities"} in k.params.values()
        assert {"dnorm": False, "key": "connectivities"} in k.params.values()
        assert {
            "softmax_scale": 4,
            "mode": "deterministic",
            "seed": 42,
            "scheme": str(CosineScheme()),
        } in k.params.values()


class TestKernel:
    def test_precomputed_not_array(self):
        with pytest.raises(TypeError):
            _ = PrecomputedKernel([[1, 0], [0, 1]])

    def test_precomputed_not_square(self):
        with pytest.raises(ValueError):
            _ = PrecomputedKernel(np.random.normal(size=(10, 9)))

    def test_precomputed_not_a_transition_matrix(self):
        mat = random_transition_matrix(100)
        mat[0, 0] = 0xDEADBEEF
        with pytest.raises(ValueError):
            _ = PrecomputedKernel(mat)

    def test_precomputed_from_kernel_no_transition(self, adata: AnnData):
        vk = VelocityKernel(adata)

        with pytest.raises(ValueError):
            PrecomputedKernel(vk)

    @pytest.mark.parametrize(
        "clazz",
        [
            ConnectivityKernel,
            VelocityKernel,
            PseudotimeKernel,
            CytoTRACEKernel,
            PrecomputedKernel,
        ],
    )
    @pytest.mark.parametrize("key_added", [None, "foo"])
    def test_kernel_reads_correct_connectivities(
        self, adata: AnnData, key_added: Optional[str], clazz: type
    ):
        del adata.uns["neighbors"]
        del adata.obsp["connectivities"]
        del adata.obsp["distances"]

        sc.pp.neighbors(adata, key_added=key_added)
        kwargs = {"adata": adata, "conn_key": key_added}

        if clazz == PseudotimeKernel:
            kwargs["time_key"] = "latent_time"
        elif clazz == PrecomputedKernel:
            adata.obsp["foo"] = np.eye(adata.n_obs)
            kwargs["transition_matrix"] = "foo"
        conn = (
            adata.obsp["connectivities"]
            if key_added is None
            else adata.obsp[f"{key_added}_connectivities"]
        )

        k = clazz(**kwargs)

        if isinstance(k, PrecomputedKernel):
            assert k._conn is None
        else:
            np.testing.assert_array_equal(k._conn.A, conn.A)

    def test_precomputed_from_kernel(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(
            mode="stochastic",
            softmax_scale=4,
        )

        pk = PrecomputedKernel(vk)
        pk.write_to_adata()

        assert pk.adata is vk.adata
        assert pk._origin == str(vk).strip("~<>")
        assert pk.params is not vk.params
        assert pk.params == vk.params
        assert pk.transition_matrix is not vk.transition_matrix
        np.testing.assert_array_equal(pk.transition_matrix.A, vk.transition_matrix.A)

    def test_precomputed_no_adata(self):
        pk = PrecomputedKernel(random_transition_matrix(50))
        pk.write_to_adata()

        assert isinstance(pk.adata, AnnData)
        assert pk._origin == "'array'"
        assert pk.adata.shape == (50, 1)
        assert pk.adata.obs.shape == (50, 0)
        assert pk.adata.var.shape == (1, 0)
        assert "T_fwd_params" in pk.adata.uns.keys()
        assert pk.adata.uns["T_fwd_params"] == {"params": pk.params}
        np.testing.assert_array_equal(
            pk.adata.obsp["T_fwd"].toarray(), pk.transition_matrix.toarray()
        )

    def test_precomputed_different_adata(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(
            mode="stochastic", softmax_scale=4
        )
        bdata = adata.copy()

        pk = PrecomputedKernel(vk, adata=bdata)

        assert pk.adata is adata
        assert pk.adata is vk.adata
        assert pk.adata is not bdata

    def test_precomputed_adata_origin(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(
            mode="stochastic", softmax_scale=4
        )
        vk.write_to_adata("foo")

        pk = PrecomputedKernel("foo", adata=adata)

        assert pk._origin == "adata.obsp['foo']"

    def test_precomputed_adata(self, adata: AnnData):
        pk = PrecomputedKernel(random_transition_matrix(adata.n_obs), adata=adata)

        assert pk.adata is adata

    def test_precomputed_transition_matrix(self, adata: AnnData):
        mat = random_transition_matrix(adata.n_obs)
        pk = PrecomputedKernel(mat)

        np.testing.assert_array_equal(mat, pk.transition_matrix.toarray())

    def test_precomputed_sum(self, adata: AnnData):
        mat = random_transition_matrix(adata.n_obs)
        pk = PrecomputedKernel(mat)
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)

        expected = (0.5 * vk.transition_matrix) + (0.5 * pk.transition_matrix)
        actual = (pk + vk).compute_transition_matrix()

        np.testing.assert_array_almost_equal(
            expected.toarray(), actual.transition_matrix.toarray()
        )

    @pytest.mark.parametrize("dnorm", [False, True])
    @pytest.mark.parametrize("sparse", [False, True])
    def test_custom_preserves_type(self, adata: AnnData, sparse: bool, dnorm: bool):
        c = CustomKernel(adata).compute_transition_matrix(sparse=sparse, dnorm=dnorm)

        if sparse:
            assert isspmatrix_csr(c.transition_matrix)
        else:
            assert isinstance(c.transition_matrix, np.ndarray)

        assert c.transition_matrix.dtype == _dtype

    def test_write_adata(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        vk.write_to_adata()

        assert adata is vk.adata
        assert "T_fwd_params" in adata.uns.keys()
        np.testing.assert_array_equal(
            adata.obsp["T_fwd"].toarray(), vk.transition_matrix.toarray()
        )

    def test_write_adata_key(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        vk.write_to_adata(key="foo")

        assert adata is vk.adata
        assert "foo_params" in adata.uns.keys()
        np.testing.assert_array_equal(
            adata.obsp["foo"].toarray(), vk.transition_matrix.toarray()
        )

    def test_row_normalized(self, adata: AnnData):
        vk = VelocityKernel(adata)

        vk.compute_transition_matrix(density_normalize=False, softmax_scale=4)
        T = vk.transition_matrix
        np.testing.assert_allclose(T.sum(1), 1, rtol=_rtol)

    @jax_not_installed_skip
    def test_row_normalized_dense_norm_stoch(self, adata: AnnData):
        vk = VelocityKernel(adata)
        vk.compute_transition_matrix(mode="stochastic", softmax_scale=4)
        T_stoch = vk.transition_matrix

        np.testing.assert_allclose(T_stoch.sum(1), 1, rtol=_rtol)

    @jax_not_installed_skip
    def test_transition_forward_stoch_high_lvl(self, adata: AnnData):
        backward = False

        vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
            mode="stochastic",
            softmax_scale=4,
        )
        T_1 = vk.transition_matrix

        T_2 = cr.tl.transition_matrix(
            adata,
            mode="stochastic",
            backward=backward,
            softmax_scale=4,
            weight_connectivities=0,
        ).transition_matrix

        np.testing.assert_allclose(T_1.A, T_2.A, rtol=_rtol)

    # only to 15 because in kernel, if a row sums to 0, abs. states are created
    # this happens because k_thresh = frac_to_keep = 0
    @pytest.mark.parametrize("k", range(1, 15))
    def test_pseudotime_frac_to_keep(self, adata: AnnData, k: int):
        conn = _get_neighs(adata, "connectivities")
        n_neighbors = _get_neighs_params(adata)["n_neighbors"]
        pseudotime = adata.obs["latent_time"]
        k_thresh = max(0, min(int(np.floor(n_neighbors / k)) - 1, 30))

        conn_biased = bias_knn(conn.copy(), pseudotime, n_neighbors, k=k)
        T_1 = _normalize(conn_biased)

        pk = PseudotimeKernel(adata, time_key="latent_time").compute_transition_matrix(
            frac_to_keep=k_thresh / float(n_neighbors),
            threshold_scheme="hard",
        )
        T_2 = pk.transition_matrix

        np.testing.assert_allclose(T_1.A, T_2.A, rtol=_rtol)

    def test_pseudotime_parallelize(self, adata: AnnData):
        pk1 = PseudotimeKernel(adata, time_key="latent_time").compute_transition_matrix(
            n_jobs=None
        )
        pk2 = PseudotimeKernel(adata, time_key="latent_time").compute_transition_matrix(
            n_jobs=2
        )

        np.testing.assert_allclose(
            pk1.transition_matrix.A, pk2.transition_matrix.A, rtol=_rtol
        )

    def test_pseudotime_inverse(self, adata: AnnData):
        pk = PseudotimeKernel(adata, time_key="latent_time")
        pt = pk.pseudotime.copy()

        pk_inv = ~pk

        assert pk_inv is pk
        assert pk_inv.backward
        np.testing.assert_allclose(pt, 1 - pk_inv.pseudotime)

    def test_pseudotime_differ_dense_norm(self, adata: AnnData):
        conn = _get_neighs(adata, "connectivities")
        n_neighbors = _get_neighs_params(adata)["n_neighbors"]
        pseudotime = adata.obs["latent_time"]

        conn_biased = bias_knn(conn, pseudotime, n_neighbors)
        T_1 = density_normalization(conn_biased, conn)
        T_1 = _normalize(T_1)

        pk = PseudotimeKernel(adata, time_key="latent_time").compute_transition_matrix()
        T_2 = pk.transition_matrix

        assert not np.allclose(T_1.A, T_2.A, rtol=_rtol)

    def test_manual_combination_det(self, adata: AnnData):
        mode = "deterministic"

        vk = VelocityKernel(adata).compute_transition_matrix(mode=mode, softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()

        T_vk = vk.transition_matrix
        T_ck = ck.transition_matrix
        T_comb_manual = 0.8 * T_vk + 0.2 * T_ck

        comb_kernel = 0.8 * vk + 0.2 * ck
        T_comb_kernel = comb_kernel.transition_matrix

        np.testing.assert_allclose(T_comb_kernel.A, T_comb_manual.A, rtol=_rtol)

    @jax_not_installed_skip
    def test_manual_combination_stoch(self, adata: AnnData):
        mode = "stochastic"

        vk = VelocityKernel(adata).compute_transition_matrix(mode=mode, softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()

        T_vk = vk.transition_matrix
        T_ck = ck.transition_matrix
        T_comb_manual = 0.8 * T_vk + 0.2 * T_ck

        comb_kernel = 0.8 * vk + 0.2 * ck
        T_comb_kernel = comb_kernel.transition_matrix

        np.testing.assert_allclose(T_comb_kernel.A, T_comb_manual.A, rtol=_rtol)

    def test_manual_combination_sam(self, adata: AnnData):
        mode = "sampling"

        vk = VelocityKernel(adata).compute_transition_matrix(mode=mode, softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix()

        T_vk = vk.transition_matrix
        T_ck = ck.transition_matrix
        T_comb_manual = 0.8 * T_vk + 0.2 * T_ck

        comb_kernel = 0.8 * vk + 0.2 * ck
        T_comb_kernel = comb_kernel.transition_matrix

        np.testing.assert_allclose(T_comb_kernel.A, T_comb_manual.A, rtol=_rtol)

    def test_manual_combination_no_precomputed(self, adata: AnnData):
        density_normalize = False

        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix(
            density_normalize=density_normalize
        )

        T_vk = vk.transition_matrix
        T_ck = ck.transition_matrix
        T_comb_manual = 0.8 * T_vk + 0.2 * T_ck

        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        ck = ConnectivityKernel(adata).compute_transition_matrix(
            density_normalize=density_normalize
        )
        comb_kernel = 0.8 * vk + 0.2 * ck
        comb_kernel.compute_transition_matrix()
        T_comb_kernel = comb_kernel.transition_matrix

        np.testing.assert_allclose(T_comb_manual.A, T_comb_kernel.A, rtol=_rtol)

    def test_manual_combination_backward(self, adata: AnnData):
        backward, density_normalize = True, False

        vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
            softmax_scale=4
        )

        ck = ConnectivityKernel(adata, backward=backward).compute_transition_matrix(
            density_normalize=density_normalize
        )

        T_vk = vk.transition_matrix
        T_ck = ck.transition_matrix
        T_comb_manual = 0.8 * T_vk + 0.2 * T_ck

        comb_kernel = 0.8 * vk + 0.2 * ck
        T_comb_kernel = comb_kernel.transition_matrix

        np.testing.assert_allclose(T_comb_manual.A, T_comb_kernel.A, rtol=_rtol)

    def test_manual_combination_backward_dense_norm(self, adata: AnnData):
        backward, density_normalize = True, True

        vk = VelocityKernel(adata, backward=backward).compute_transition_matrix(
            softmax_scale=4
        )

        ck = ConnectivityKernel(adata, backward=backward).compute_transition_matrix(
            density_normalize=density_normalize
        )

        T_vk = vk.transition_matrix
        T_ck = ck.transition_matrix
        T_comb_manual = 0.8 * T_vk + 0.2 * T_ck

        comb_kernel = 0.8 * vk + 0.2 * ck
        T_comb_kernel = comb_kernel.transition_matrix

        np.testing.assert_allclose(T_comb_manual.A, T_comb_kernel.A, rtol=_rtol)

    def compare_with_scanpy_density_normalize(self, adata: AnnData):
        # check whether cellrank's transition matrix matches scanpy's
        density_normalize = True
        ck = ConnectivityKernel(adata).compute_transition_matrix(
            density_normalize=density_normalize
        )
        T_cr = ck.transition_matrix

        neigh = Neighbors(adata)
        neigh.compute_transitions(density_normalize=density_normalize)
        T_sc = neigh.transitions

        # check whether these are the same while leaving them sparse
        assert T_sc.shape == T_cr.shape
        assert len(T_sc.indices) == len(T_cr.indices)
        assert np.allclose((T_cr - T_sc).data, 0)

    def compare_with_scanpy(self, adata: AnnData):
        # check whether cellrank's transition matrix matches scanpy's
        density_normalize = False
        ck = ConnectivityKernel(adata).compute_transition_matrix(
            density_normalize=density_normalize
        )
        T_cr = ck.transition_matrix

        neigh = Neighbors(adata)
        neigh.compute_transitions(density_normalize=density_normalize)
        T_sc = neigh.transitions

        # check whether these are the same while leaving them sparse
        assert T_sc.shape == T_cr.shape
        assert len(T_sc.indices) == len(T_cr.indices)
        assert np.allclose((T_cr - T_sc).data, 0)

    def test_connectivities_key_kernel(self, adata: AnnData):
        key = "foobar"
        assert key not in adata.obsp
        adata.obsp[key] = np.eye(adata.n_obs)

        ck = ConnectivityKernel(adata, conn_key=key).compute_transition_matrix()
        T_cr = ck.transition_matrix

        assert key == ck.params["key"]
        np.testing.assert_array_equal(T_cr, adata.obsp[key])

        del adata.obsp[key]


class TestKernelAddition:
    def test_simple_addition(self, adata: AnnData):
        vk, ck = create_kernels(adata)  # diagonal + upper diag

        k = (vk + ck).compute_transition_matrix()
        expected = np.eye(adata.n_obs) * 0.75 + np.eye(adata.n_obs, k=1) * 0.25
        expected[-1, -1] = 1

        np.testing.assert_allclose(k.transition_matrix.A, expected)

    def test_addtion_with_constant(self, adata: AnnData):
        vk, ck = create_kernels(adata)  # diagonal + upper diag

        k = (2 * vk + 3 * ck).compute_transition_matrix()
        expected = (
            np.eye(adata.n_obs) * (2 / 5)
            + np.eye(adata.n_obs) * (3 / 5) * 0.5
            + np.eye(adata.n_obs, k=1) * (3 / 5) * 0.5
        )
        expected[-1, -1] = 1

        np.testing.assert_allclose(k.transition_matrix.A, expected)

    def test_addition_3_kernels(self, adata: AnnData):
        vk, ck = create_kernels(adata)  # diagonal + upper diag
        vk1 = VelocityKernel(adata)
        vk1._transition_matrix = np.eye(adata.n_obs, k=-1) / 2 + np.eye(adata.n_obs) / 2
        vk1._transition_matrix[0, 0] = 1
        np.testing.assert_allclose(
            np.sum(ck._transition_matrix, axis=1), 1
        )  # sanity check

        k = (vk + ck + vk1).compute_transition_matrix()
        expected = (
            np.eye(adata.n_obs) * (1 / 3 + 1 / 6 + 1 / 6)
            + np.eye(adata._n_obs, k=1) * 1 / 6
            + np.eye(adata.n_obs, k=-1) * 1 / 6
        )
        expected[0, 0] = expected[-1, -1] = 2 / 3 + 1 / 3 * 0.5
        expected[0, 1] = expected[-1, -2] = 1 - expected[0, 0]

        np.testing.assert_allclose(k.transition_matrix.A, expected)

    def test_addition_adaptive(self, adata: AnnData):
        adata.uns["velocity_variances"] = vv = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        adata.uns["connectivity_variances"] = cv = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        vk, ck = create_kernels(
            adata,
            velocity_variances="velocity_variances",
            connectivity_variances="connectivity_variances",
        )

        k = vk ^ ck
        expected = _normalize(
            0.5 * vv * vk.transition_matrix + 0.5 * cv * ck.transition_matrix
        )

        np.testing.assert_allclose(k.transition_matrix.A, expected)

    def test_addition_adaptive_constants(self, adata: AnnData):
        a, b = np.random.uniform(0, 10, 2)
        s = a + b
        adata.uns["velocity_variances"] = vv = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        adata.uns["connectivity_variances"] = cv = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        vk, ck = create_kernels(
            adata,
            velocity_variances="velocity_variances",
            connectivity_variances="connectivity_variances",
        )

        k = a * vk ^ b * ck
        expected = _normalize(
            a / s * vv * vk.transition_matrix + b / s * cv * ck.transition_matrix
        )

        np.testing.assert_allclose(k.transition_matrix.A, expected)

    def test_addition_adaptive_wrong_variances(self, adata: AnnData):
        a, b = np.random.uniform(0, 10, 2)
        s = a + b
        adata.uns["velocity_variances"] = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        adata.uns["connectivity_variances"] = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        vk, ck = create_kernels(
            adata,
            velocity_variances="velocity_variances",
            connectivity_variances="connectivity_variances",
        )

        k = a * vk ^ b * ck
        expected = _normalize(
            a / s * vk.transition_matrix + b / s * ck.transition_matrix
        )

        assert not np.allclose(k.transition_matrix.A, expected.A)

    def test_addition_adaptive_4_kernels(self, adata: AnnData):
        a, b, c, d = np.random.uniform(0, 10, 4)
        s = a + b + c + d
        adata.uns["velocity_variances"] = vv = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        adata.uns["connectivity_variances"] = cv = np.random.random(
            size=(adata.n_obs, adata.n_obs)
        )
        vk, ck = create_kernels(
            adata,
            velocity_variances="velocity_variances",
            connectivity_variances="connectivity_variances",
        )
        vk1, ck1 = create_kernels(
            adata,
            velocity_variances="velocity_variances",
            connectivity_variances="connectivity_variances",
        )

        k = a * vk ^ b * ck ^ c * vk1 ^ d * ck1
        expected = _normalize(
            a / s * vv * vk.transition_matrix
            + b / s * cv * ck.transition_matrix
            + c / s * vv * vk1.transition_matrix
            + d / s * cv * ck1.transition_matrix
        )

        np.testing.assert_allclose(k.transition_matrix.A, expected)


class TestKernelCopy:
    def test_copy_simple(self, adata: AnnData):
        vk1 = VelocityKernel(adata)
        vk2 = vk1.copy()

        assert vk1 is not vk2

    def test_copy_no_adata_copy(self, adata: AnnData):
        vk1 = VelocityKernel(adata)
        vk2 = vk1.copy()

        assert vk1.adata is adata
        assert vk2.adata is adata

    def test_copy_transition_matrix(self, adata: AnnData):
        vk1 = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        vk2 = vk1.copy()

        np.testing.assert_array_equal(vk1.transition_matrix.A, vk2.transition_matrix.A)

    def test_copy_params(self, adata: AnnData):
        vk1 = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        vk2 = vk1.copy()

        assert vk1.params == vk2.params

    def test_copy_cond_num(self, adata: AnnData):
        for KernelClass in [
            VelocityKernel,
            ConnectivityKernel,
            PseudotimeKernel,
            PrecomputedKernel,
        ]:
            if KernelClass is PrecomputedKernel:
                k1 = KernelClass(
                    random_transition_matrix(adata.n_obs), compute_cond_num=True
                )
            elif KernelClass is VelocityKernel:
                k1 = KernelClass(
                    adata, compute_cond_num=True
                ).compute_transition_matrix(softmax_scale=4)
            else:
                k1 = KernelClass(
                    adata, compute_cond_num=True
                ).compute_transition_matrix()
            k2 = k1.copy()

            assert k1.condition_number == k2.condition_number

    def test_copy_velocity_kernel(self, adata: AnnData):
        vk1 = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)
        vk2 = vk1.copy()

        np.testing.assert_array_equal(vk1.transition_matrix.A, vk2.transition_matrix.A)
        np.testing.assert_array_equal(vk1.logits.A, vk2.logits.A)

        assert vk1.params == vk2.params
        assert vk1.backward == vk2.backward

    def test_copy_connectivity_kernel(self, adata: AnnData):
        ck1 = ConnectivityKernel(adata).compute_transition_matrix()
        ck2 = ck1.copy()

        np.testing.assert_array_equal(ck1.transition_matrix.A, ck2.transition_matrix.A)
        assert ck1.params == ck2.params
        assert ck1.backward == ck2.backward

    def test_copy_palantir_kernel(self, adata: AnnData):
        pk1 = PseudotimeKernel(adata).compute_transition_matrix()
        pk2 = pk1.copy()

        np.testing.assert_array_equal(pk1.transition_matrix.A, pk2.transition_matrix.A)
        assert pk1.params == pk2.params
        assert pk1.backward == pk2.backward

    def test_copy_works(self, adata: AnnData):
        ck1 = ConnectivityKernel(adata)
        ck2 = ck1.copy()
        ck1.compute_transition_matrix()

        assert (
            ck1._transition_matrix is not None
        )  # calling the property would trigger the calculation
        assert ck2._transition_matrix is None


class TestGeneral:
    def test_kernels(self, adata: AnnData):
        vk = VelocityKernel(adata)

        assert len(vk.kernels) == 1
        assert vk.kernels[0] is vk

    def test_kernels_multiple(self, adata: AnnData):
        vk = VelocityKernel(adata)
        ck = ConnectivityKernel(adata)
        v = vk + ck

        assert len(v.kernels) == 2
        assert vk in v.kernels
        assert ck in v.kernels

    def test_kernels_multiple_constant(self, adata: AnnData):
        vk = VelocityKernel(adata)
        ck = ConnectivityKernel(adata)
        v = 100 * vk + 42 * ck

        assert len(v.kernels) == 2
        assert vk in v.kernels
        assert ck in v.kernels

    def test_kernels_unique(self, adata: AnnData):
        vk = VelocityKernel(adata)
        v = vk + vk + vk + vk

        assert len(v.kernels) == 1
        assert v.kernels[0] is vk

    def test_no_comp_cond_num(self, adata: AnnData):
        vk = VelocityKernel(adata).compute_transition_matrix(softmax_scale=4)

        assert vk.condition_number is None

    def test_comp_cond_num(self, adata: AnnData):
        vk = VelocityKernel(adata, compute_cond_num=True).compute_transition_matrix(
            softmax_scale=4
        )

        assert isinstance(vk.condition_number, float)

    def test_comp_cond_num_or_policy(self, adata: AnnData):
        vk = VelocityKernel(adata, compute_cond_num=True).compute_transition_matrix(
            softmax_scale=4
        )
        ck = ConnectivityKernel(
            adata, compute_cond_num=False
        ).compute_transition_matrix()
        v = (vk + ck).compute_transition_matrix()

        assert isinstance(vk.condition_number, float)
        assert ck.condition_number is None
        assert isinstance(v.condition_number, float)


class TestTransitionProbabilities:
    def test_pearson_correlations_fwd(self, adata: AnnData):
        # test whether pearson correlations in cellrank match those from scvelo, forward case
        backward = False

        # compute pearson correlations using scvelo
        velo_graph = adata.uns["velocity_graph"] + adata.uns["velocity_graph_neg"]

        # compute pearson correlations using cellrank
        vk = VelocityKernel(adata, backward=backward)
        vk.compute_transition_matrix(mode="deterministic", softmax_scale=4)
        pearson_correlations_cr = vk.logits

        pc_r = velo_graph.copy()
        pc_r.data = np.array(pearson_correlations_cr[(velo_graph != 0)]).squeeze()

        assert np.max(np.abs((pc_r - velo_graph).data)) < _rtol

    def test_pearson_correlations_bwd(self, adata: AnnData):
        # test whether pearson correlations in cellrank match those from scvelo, backward case
        backward = True

        # compute pearson correlations using scvelo
        velo_graph = (adata.uns["velocity_graph"] + adata.uns["velocity_graph_neg"]).T

        # compute pearson correlations using cellrak
        vk = VelocityKernel(adata, backward=backward)
        vk.compute_transition_matrix(
            mode="deterministic", backward_mode="transpose", softmax_scale=4
        )
        pearson_correlations_cr = vk.logits

        pc_r = velo_graph.copy()
        pc_r.data = np.array(pearson_correlations_cr[(velo_graph != 0)]).squeeze()

        assert np.max(np.abs((pc_r - velo_graph.T).data)) < _rtol

    def test_transition_probabilities_fwd(self, adata: AnnData):
        # test whether transition probabilities in cellrank match those from scvelo, forward case
        sigma_test = 3

        # compute transition probabilities using cellrank
        vk = VelocityKernel(adata)
        vk.compute_transition_matrix(softmax_scale=sigma_test, mode="deterministic")
        T_cr = vk.transition_matrix

        pearson_correlation = vk.logits
        T_exp = np.expm1(pearson_correlation * sigma_test)
        T_exp.data += 1
        T_exp = _normalize(T_exp)

        np.testing.assert_allclose(T_exp.A, T_cr.A)  # don't use data, can be reordered

    def test_transition_probabilities_bwd(self, adata: AnnData):
        # test whether transition probabilities in cellrank match those from scvelo, backward case
        sigma_test = 3

        # compute transition probabilities using cellrank
        vk = VelocityKernel(adata, backward=True)
        vk.compute_transition_matrix(softmax_scale=sigma_test, mode="deterministic")
        T_cr = vk.transition_matrix

        pearson_correlation = vk.logits
        T_exp = np.expm1(pearson_correlation * sigma_test)
        T_exp.data += 1
        T_exp = _normalize(T_exp)

        np.testing.assert_allclose(T_exp.A, T_cr.A)  # don't use data, can be reordered

    def test_estimate_softmax_scale(self, adata: AnnData):
        vk = VelocityKernel(adata)
        vk.compute_transition_matrix(
            mode="deterministic", show_progress_bar=False, softmax_scale=None
        )

        assert isinstance(vk.params["softmax_scale"], float)


class TestMonteCarlo:
    def test_mc_and_mc_fwd_1k(self, adata: AnnData):
        vk_mc = VelocityKernel(adata, backward=False)
        vk_mc.compute_transition_matrix(
            mode="monte_carlo",
            show_progress_bar=False,
            n_samples=1000,
            n_jobs=4,
            softmax_scale=4,
        )

        vk_s = VelocityKernel(adata, backward=False)
        vk_s.compute_transition_matrix(
            mode="monte_carlo",
            show_progress_bar=False,
            n_samples=1000,
            n_jobs=4,
            softmax_scale=4,
        )

        val = np.mean(
            np.abs(vk_mc.transition_matrix.data - vk_s.transition_matrix.data)
        )
        assert val < 1e-5, val

    def test_mc_and_mc_fwd_10k(self, adata: AnnData):
        vk_mc = VelocityKernel(adata, backward=False)
        vk_mc.compute_transition_matrix(
            mode="monte_carlo",
            show_progress_bar=False,
            n_samples=10000,
            n_jobs=4,
            softmax_scale=4,
        )

        vk_s = VelocityKernel(adata, backward=False)
        vk_s.compute_transition_matrix(
            mode="monte_carlo",
            show_progress_bar=False,
            n_samples=10000,
            n_jobs=4,
            softmax_scale=4,
        )

        val = np.mean(
            np.abs(vk_mc.transition_matrix.data - vk_s.transition_matrix.data)
        )
        assert val < 1e-6, val

    @jax_not_installed_skip
    def test_mc_and_second_order_fwd(self, adata: AnnData):
        vk_mc = VelocityKernel(adata, backward=False)
        vk_mc.compute_transition_matrix(
            mode="monte_carlo",
            show_progress_bar=False,
            n_samples=1000,
            n_jobs=4,
            softmax_scale=4,
        )

        vk_s = VelocityKernel(adata, backward=False)
        vk_s.compute_transition_matrix(
            mode="stochastic", show_progress_bar=False, n_jobs=4, softmax_scale=4
        )

        val = np.mean(
            np.abs(vk_mc.transition_matrix.data - vk_s.transition_matrix.data)
        )
        assert val < 1e-3, val


class TestVelocityScheme:
    def test_invalid_string_key(self, adata: AnnData):
        vk = VelocityKernel(adata)
        with pytest.raises(ValueError):
            vk.compute_transition_matrix(scheme="foobar")

    def test_not_callable(self, adata: AnnData):
        vk = VelocityKernel(adata)
        with pytest.raises(
            TypeError, match="Expected `scheme` to be a function, found"
        ):
            vk.compute_transition_matrix(scheme=1311)

    def test_custom_function_not_sum_to_1(self, adata: AnnData):
        vk = VelocityKernel(adata)
        with pytest.raises(ValueError, match=r"Matrix is not row-stochastic."):
            vk.compute_transition_matrix(scheme=InvalidFuncProbs())

    def test_custom_function_invalid_hessian(self, adata: AnnData):
        vk = VelocityKernel(adata)
        with pytest.raises(ValueError, match=r"Expected full Hessian matrix"):
            vk.compute_transition_matrix(
                mode="stochastic", scheme=InvalidFuncHessianShape(), softmax_scale=4
            )

    @pytest.mark.parametrize("backward", [True, False])
    def test_implementations_differ(self, adata: AnnData, backward: bool):
        vk_dot = VelocityKernel(adata, backward=backward)
        vk_dot.compute_transition_matrix(
            mode="deterministic", softmax_scale=4, scheme="dot_product"
        )
        vk_cos = VelocityKernel(adata, backward=backward)
        vk_cos.compute_transition_matrix(
            mode="deterministic", softmax_scale=4, scheme="cosine"
        )
        vk_cor = VelocityKernel(adata, backward=backward)
        vk_cor.compute_transition_matrix(
            mode="deterministic", softmax_scale=4, scheme="correlation"
        )

        np.testing.assert_allclose(vk_dot.transition_matrix.sum(1), 1.0)
        np.testing.assert_allclose(vk_cor.transition_matrix.sum(1), 1.0)
        np.testing.assert_allclose(vk_cor.transition_matrix.sum(1), 1.0)

        assert not np.allclose(vk_dot.transition_matrix.A, vk_cos.transition_matrix.A)
        assert not np.allclose(vk_cos.transition_matrix.A, vk_cor.transition_matrix.A)
        assert not np.allclose(vk_cor.transition_matrix.A, vk_dot.transition_matrix.A)

    @pytest.mark.parametrize(
        "key,fn",
        zip(
            ["dot_product", "cosine", "correlation"],
            [
                cr.tl.kernels.DotProductScheme(),
                cr.tl.kernels.CosineScheme(),
                cr.tl.kernels.CorrelationScheme(),
            ],
        ),
    )
    def test_function_and_string_key(self, adata: AnnData, key: str, fn: Callable):
        vk_k = VelocityKernel(adata)
        vk_fn = VelocityKernel(adata)

        vk_k.compute_transition_matrix(
            mode="deterministic", softmax_scale=4, scheme=key
        )
        vk_fn.compute_transition_matrix(
            mode="deterministic", softmax_scale=4, scheme=fn
        )

        np.testing.assert_allclose(vk_k.transition_matrix.A, vk_fn.transition_matrix.A)

    @pytest.mark.parametrize("backward", [True, False])
    def test_custom_function(self, adata: AnnData, backward: bool):
        vk = VelocityKernel(adata, backward=backward)
        vk.compute_transition_matrix(
            mode="deterministic", softmax_scale=4, scheme=CustomFuncHessian()
        )

        assert vk.params["scheme"] == str(CustomFuncHessian())

    def test_custom_function_stochastic_no_hessian(self, adata: AnnData):
        vk = VelocityKernel(adata)
        vk.compute_transition_matrix(
            mode="stochastic", scheme=CustomFunc(), softmax_scale=4, n_samples=10
        )

        assert vk.params["mode"] == "monte_carlo"
        assert vk.params["scheme"] == str(CustomFunc())


class TestComputeProjection:
    def test_no_transition_matrix(self, adata: AnnData):
        with pytest.raises(RuntimeError, match=r"Compute transition matrix first as"):
            cr.tl.kernels.ConnectivityKernel(adata).compute_projection()

    def test_no_basis(self, adata: AnnData):
        ck = cr.tl.kernels.ConnectivityKernel(adata).compute_transition_matrix()
        with pytest.raises(KeyError, match=r"Unable to find a basis in"):
            ck.compute_projection(basis="foo")

    def test_basis_prefix(self, adata: AnnData):
        ck = cr.tl.kernels.ConnectivityKernel(adata).compute_transition_matrix()
        ck.compute_projection(basis="X_umap")

    @pytest.mark.parametrize("write_first", [True, False])
    def test_write_to_adata(self, adata: AnnData, write_first: bool):
        ck = cr.tl.kernels.ConnectivityKernel(adata).compute_transition_matrix()
        if write_first:
            ck.write_to_adata()
            ck.compute_projection(basis="umap")
        else:
            ck.compute_projection(basis="umap")
            ck.write_to_adata()

        assert adata.uns[_transition(ck._direction) + "_params"] == {
            "params": ck.params,
            "embeddings": ["umap"],
        }

    @pytest.mark.parametrize("key_added", [None, "foo"])
    def test_key_added(self, adata: AnnData, key_added: Optional[str]):
        ck = cr.tl.kernels.ConnectivityKernel(adata).compute_transition_matrix()
        ck.compute_projection(basis="umap", copy=False, key_added=key_added)

        key = key_added if key_added is not None else _transition(ck._direction)
        ukey = f"{key}_params"
        key = f"{key}_umap"

        assert adata.uns[ukey] == {"embeddings": ["umap"]}
        np.testing.assert_array_equal(adata.obsm[key].shape, adata.obsm["X_umap"].shape)

    @pytest.mark.parametrize("copy", [True, False])
    def test_copy(self, adata: AnnData, copy: bool):
        ck = cr.tl.kernels.ConnectivityKernel(adata).compute_transition_matrix()
        res = ck.compute_projection(basis="umap", copy=copy)

        if copy:
            assert isinstance(res, np.ndarray)
            np.testing.assert_array_equal(res.shape, adata.obsm["X_umap"].shape)
        else:
            assert res is None
            key = _transition(ck._direction) + "_umap"
            np.testing.assert_array_equal(
                adata.obsm[key].shape, adata.obsm["X_umap"].shape
            )

    def test_nan_in_embedding(self, adata: AnnData):
        adata.obsm["X_umap"][-1] = np.nan

        ck = cr.tl.kernels.ConnectivityKernel(adata).compute_transition_matrix()
        res = ck.compute_projection(basis="umap", copy=True)

        assert not np.all(np.isnan(res))
        assert np.all(np.isnan(res[-1, :]))


class TestPseudotimeKernelScheme:
    def test_invalid_scheme(self, adata: AnnData):
        pk = PseudotimeKernel(adata)
        with pytest.raises(ValueError, match="foo"):
            pk.compute_transition_matrix(threshold_scheme="foo")

    def test_invalid_custom_scheme(self, adata: AnnData):
        pk = PseudotimeKernel(adata)
        with pytest.raises(ValueError, match="Expected row of shape"):
            pk.compute_transition_matrix(
                threshold_scheme=lambda cpt, npt, ndist: np.ones(
                    (len(ndist) - 1), dtype=np.float64
                ),
            )

    def test_custom_scheme(self, adata: AnnData):
        pk = PseudotimeKernel(adata)
        pk.compute_transition_matrix(
            threshold_scheme=lambda cpt, npt, ndist: np.ones(
                (len(ndist)), dtype=np.float64
            ),
        )

        np.testing.assert_allclose(pk.transition_matrix.sum(1), 1.0)
        for row in pk.transition_matrix:
            np.testing.assert_allclose(row.data, 1 / len(row.data))

    @pytest.mark.parametrize("scheme", ["hard", "soft"])
    def test_scheme(self, adata: AnnData, scheme: str):
        pk = PseudotimeKernel(adata)
        pk.compute_transition_matrix(
            threshold_scheme=scheme, frac_to_keep=0.3, b=10, nu=0.5
        )

        np.testing.assert_allclose(pk.transition_matrix.sum(1), 1.0)
        assert pk.params["scheme"] == scheme
        if scheme == "hard":
            assert pk.params["frac_to_keep"] == 0.3
            assert "b" not in pk.params
            assert "nu" not in pk.params
        elif scheme == "soft":
            assert pk.params["b"] == 10
            assert pk.params["nu"] == 0.5
            assert "k" not in pk.params


class TestCytoTRACEKernel:
    @pytest.mark.parametrize("layer", ["X", "Ms", "foo"])
    def test_layer(self, adata: AnnData, layer: str):
        if layer == "foo":
            with pytest.raises(KeyError, match=layer):
                _ = CytoTRACEKernel(adata, layer=layer)
        else:
            _ = CytoTRACEKernel(adata, layer=layer)
            assert adata.uns[_ct("params")]["layer"] == layer

    @pytest.mark.parametrize("agg", list(CytoTRACEAggregation))
    def test_aggregation(self, adata: AnnData, agg: CytoTRACEAggregation):
        _ = CytoTRACEKernel(adata, aggregation=agg)
        assert adata.uns[_ct("params")]["aggregation"] == agg.s

    @pytest.mark.parametrize("use_raw", [False, True])
    def test_raw(self, adata: AnnData, use_raw: bool):
        _ = CytoTRACEKernel(adata, use_raw=use_raw)
        assert adata.uns[_ct("params")]["use_raw"] == (
            adata.raw.n_vars == adata.n_vars if use_raw else False
        )

    def test_correct_class(self, adata: AnnData):
        k = CytoTRACEKernel(adata)
        assert isinstance(k, PseudotimeKernel)
        assert k._time_key == _ct("pseudotime")

    def test_writes_params(self, adata: AnnData):
        k = CytoTRACEKernel(adata, use_raw=False, layer="X", aggregation="mean")

        assert adata.uns[_ct("params")] == {
            "layer": "X",
            "aggregation": "mean",
            "use_raw": False,
        }

        assert np.all(adata.var[_ct("gene_corr")] <= 1.0)
        assert np.all(-1 <= adata.var[_ct("gene_corr")])
        assert is_bool_dtype(adata.var[_ct("correlates")])
        assert adata.var[_ct("correlates")].sum() == min(200, adata.n_vars)

        assert _ct("score") in adata.obs
        assert _ct("pseudotime") in adata.obs
        assert _ct("num_exp_genes") in adata.obs
        assert is_integer_dtype(adata.obs[_ct("num_exp_genes")])
        np.testing.assert_array_equal(k.pseudotime, adata.obs[_ct("pseudotime")].values)
        np.testing.assert_array_equal(k.pseudotime.min(), 0.0)
        np.testing.assert_array_equal(k.pseudotime.max(), 1.0)

    def test_compute_transition_matrix(self, adata: AnnData):
        k = CytoTRACEKernel(adata, use_raw=False, layer="X", aggregation="mean")
        k.compute_transition_matrix()

        np.testing.assert_allclose(k.transition_matrix.sum(1), 1.0)

    def test_inversion(self, adata: AnnData):
        k = ~CytoTRACEKernel(adata, use_raw=False, layer="X", aggregation="mean")

        pt = adata.obs[_ct("pseudotime")].values
        np.testing.assert_array_equal(np.max(pt) - pt, k.pseudotime)

    def test_regression_score(self, adata_large: AnnData):
        _ = CytoTRACEKernel(adata_large, use_raw=False)

        ct_score_expected = adata_large.obs["ct_score_ground_truth"].values
        ct_score_actual = adata_large.obs[_ct("score")].values

        np.testing.assert_array_equal(ct_score_actual, ct_score_expected)

    def test_regression_gene_corr(self, adata_large: AnnData):
        _ = CytoTRACEKernel(adata_large, use_raw=False)

        gene_corr_expected = adata_large.var["gene_corr_ground_truth"].values
        gene_corr_actual = adata_large.var[_ct("gene_corr")].values

        np.testing.assert_array_equal(gene_corr_actual, gene_corr_expected)


class TestSingleFlow:
    def test_no_transition_matrix(self, kernel: Kernel):
        kernel._transition_matrix = None
        with pytest.raises(RuntimeError, match=r"Compute transition matrix first as"):
            kernel.plot_single_flow("Astrocytes", "clusters", "age(days)")

    def test_invalid_cluster_key(self, kernel: Kernel):
        with pytest.raises(KeyError, match=r"Unable to find clusters in"):
            kernel.plot_single_flow("Astrocytes", "foo", "age(days)")

    def test_invalid_source_cluster(self, kernel: Kernel):
        with pytest.raises(ValueError, match=r"Invalid source cluster"):
            kernel.plot_single_flow("foo", "clusters", "age(days)")

    def test_too_few_invalid_clusters(self, kernel: Kernel):
        with pytest.raises(ValueError, match=r"Expected at least `2` clusters"):
            kernel.plot_single_flow(
                "Astrocytes", "clusters", "age(days)", clusters=["foo", "bar", "baz"]
            )

    def test_all_invalid_clusters(self, kernel: Kernel):
        with pytest.raises(ValueError, match=r"No valid clusters have been selected."):
            kernel.plot_single_flow(
                "quux", "clusters", "age(days)", clusters=["foo", "bar", "baz"]
            )

    def test_invalid_time_key(self, kernel: Kernel):
        with pytest.raises(
            KeyError, match=r"Unable to find data in `adata.obs\['foo'\]`."
        ):
            kernel.plot_single_flow("Astrocytes", "clusters", "foo")

    def test_too_few_valid_timepoints(self, kernel: Kernel):
        with pytest.raises(ValueError, match=r"Expected at least `2` time points"):
            kernel.plot_single_flow(
                "Astrocytes", "clusters", "age(days)", time_points=["35"]
            )

    def test_all_invalid_times(self, kernel: Kernel):
        with pytest.raises(ValueError, match=r"No valid time points"):
            kernel.plot_single_flow(
                "Astrocytes", "clusters", "age(days)", time_points=[0, 1, 2]
            )

    def test_time_key_cannot_be_coerced_to_numeric(self, kernel: Kernel):
        with pytest.raises(TypeError, match=r"Unable to convert .* to `float`."):
            kernel.plot_single_flow("Astrocytes", "clusters", "clusters")

    def test_remove_empty_clusters_none_remain(self, kernel: Kernel):
        with pytest.raises(ValueError, match=r"After removing clusters with no"):
            kernel.plot_single_flow(
                "Astrocytes",
                "clusters",
                "age(days)",
                min_flow=np.inf,
                remove_empty_clusters=True,
            )



import os
from pathlib import Path

from setuptools import setup, find_packages

try:
    from cellrank import __email__, __author__, __version__, __maintainer__
except ImportError:
    __author__ = "Marius Lange, Michal Klein"
    __maintainer__ = "Marius Lange, Michal Klein"
    __email__ = "info@cellrank.org"
    __version__ = "1.4.0"


if __name__ == "__main__":
    setup(
        name="cellrank",
        use_scm_version=True,
        setup_requires=["setuptools_scm"],
        author=__author__,
        author_email=__email__,
        maintainer=__maintainer__,
        maintainer_email=__email__,
        version=__version__,
        description=Path("README.rst").read_text("utf-8").split("\n")[2],
        long_description=Path("README.rst").read_text("utf-8"),
        url="https://github.com/theislab/cellrank",
        project_urls={
            "Documentation": "https://cellrank.readthedocs.io/en/stable",
            "Source Code": "https://github.com/theislab/cellrank",
        },
        download_url="https://github.com/theislab/cellrank",
        license="BSD",
        install_requires=list(
            map(
                str.strip,
                open(os.path.abspath("requirements.txt")).read().splitlines(),
            )
        ),
        extras_require=dict(
            # `POT` is not requirement of `statot`
            external=["statot>=0.0.14", "POT", "wot"],
            krylov=["pygpcca[slepc]"],
            test=[
                "pytest>=6.1.1",
                "pytest-mock>=3.5.1",
                "pytest-xdist>=2.1.0",
                "pytest-cov",
                "Pillow",
                "filelock",
                "python-igraph",
                "louvain>=0.6,!=0.6.2",
                "leidenalg",
                "bezier",
                "jax",
                "jaxlib",
            ],
            docs=[
                r
                for r in map(
                    str.strip,
                    open(os.path.abspath("docs/requirements.txt")).read().splitlines(),
                )
                if "requirements.txt" not in r
            ],
            dev=["pre-commit>=2.9.3", "tox>=3.23.0"],
        ),
        zip_safe=False,
        packages=find_packages(),
        python_requires=">=3.6",
        platforms=["Linux", "MacOs", "Windows"],
        keywords=sorted(
            [
                "single-cell",
                "bio-informatics",
                "single-cell",
                "RNA velocity",
                "Markov chain",
                "GPCCA",
            ]
        ),
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Natural Language :: English",
            "Framework :: Jupyter",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "Typing :: Typed",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Scientific/Engineering :: Visualization",
        ],
    )
